"""
Construct Pauli operators and Hamiltonians.
"""
from collections import Counter
from typing import List, Union, Tuple
from sympy import Symbol
import numpy as np
from scipy.sparse import kron as sparse_kron
from functools import reduce
from scipy.sparse import csr_matrix

Identity = np.array(([1, 0], [0, 1]), dtype=complex)
PauliX = np.array(([0, 1], [1, 0]), dtype=complex)
PauliY = np.array(([0, -1j], [1j, 0]), dtype=complex)
PauliZ = np.array(([1, 0], [0, -1]), dtype=complex)

PAULIS_SET = set("XYZI")

PAULI_MULT_RULES = {
    "XX": "I",
    "YY": "I",
    "ZZ": "I",
    "XY": "Z",
    "YX": "Z",
    "XZ": "Y",
    "ZX": "Y",
    "YZ": "X",
    "ZY": "X",
}
PAULI_MULT_RULES.update({f"I{op}": op for op in PAULIS_SET})
PAULI_MULT_RULES.update({f"{op}I": op for op in PAULIS_SET})

PAULI_MAPPERS = {
    "X": PauliX,
    "Y": PauliY,
    "Z": PauliZ,
    "I": Identity,
    "iX": 1j * PauliX,
    "iY": 1j * PauliY,
    "iZ": 1j * PauliZ,
    "-iX": -1j * PauliX,
    "-iY": -1j * PauliY,
    "-iZ": -1j * PauliZ,
}

PAULI_PHASE_MAPPERS = {
    "XX": 1,
    "YY": 1,
    "ZZ": 1,
    "XY": 1j,
    "ZX": 1j,
    "YZ": 1j,
    "ZY": -1j,
    "XZ": -1j,
    "YX": -1j,
    "X": 1,
    "Y": 1,
    "Z": 1,
    "I": 1,
}
PAULI_PHASE_MAPPERS.update({f"I{op}": 1 for op in PAULIS_SET})
PAULI_PHASE_MAPPERS.update({f"{op}I": 1 for op in PAULIS_SET})


class PauliOp:
    """
    Pauli operator class to handle Pauli operators.

    Attributes
    ----------
    qubit_indices: `Tuple[int]`
        Indices of each spin in the Pauli operators.

    pauli_str: `str`
        Names of each Pauli opeator acting on each spin.

    phase: `complex`
        Overall complex phase of the Pauli operators.

    is_trivial

    matrix
    """

    def __init__(self, pauli_str: str, qubit_indices: Tuple[int]):
        """
        Initialize the Pauli Operator object.

        Parameters
        ----------
        pauli_str: `str`
                The Pauli operator basis string.
        qubit_indices: `Tuple[int]`
                The qubits on which the Pauli operates.

        Attributes
        ----------
        pauli_str: `str`
                The Pauli operator basis string.
        qubit_indices: `Tuple[int]`
                The qubits on which the Pauli operates.
        phase: `complex`
                The phase of the Pauli operator.
        """
        # Ensure number of indices matches number of declared operator names
        assert len(pauli_str) == len(
            qubit_indices
        ), "Each Pauli operator must have a unique qubit index"

        # Simplify Pauli string if multiple operators act on the same qubit
        pauli_str, qubit_indices, phase = self._simplify(pauli_str, qubit_indices)

        # Sort indices and strings
        self.qubit_indices, self.pauli_str = self._sort_pauli_op(
            qubit_indices, pauli_str
        )

        # Store phase accumulated from simplification
        self.phase = phase

    def __hash__(self) -> int:
        return hash((self.qubit_indices, self.pauli_str, self.phase))

    @staticmethod
    def _sort_pauli_op(qubit_indices: Tuple[int], pauli_str: str):
        """
        Sort the Pauli Operator in increasing order of qubit indices.

        Parameters
        ----------
        qubit_indices: `Tuple[int]`
                The qubit indices of the Pauli Operator.
        pauli_str: `str`
                The Pauli Operator basis string.

        Returns
        -------
        sorted_qubit_indices: `Tuple[int]`
                The sorted qubit indices in increasing order.
        sorted_pauli_str: `str`
                The sorted Pauli Operator basis string.

        Examples
        --------
        >>> PauliOp('YZX',(3,1,2)) -> PauliOp('ZXY',(1,2,3)) with appropriate phase.
        """
        # Initialize sorted pauli string and indices
        sorted_pauli_str = ""
        sorted_qubit_indices = []

        # Sorting
        for index, string in sorted(zip(qubit_indices, pauli_str)):
            # Ensure string is a valid Pauli operator
            if string not in PAULIS_SET:
                raise ValueError(
                    f"{string} is not a valid Pauli. Please choose from the set {PAULIS_SET}"
                )

            # Store sorted indices and strings
            sorted_qubit_indices.append(index)
            sorted_pauli_str += string

        sorted_qubit_indices = tuple(sorted_qubit_indices)

        return sorted_qubit_indices, sorted_pauli_str

    @staticmethod
    def _simplify(pauli_str: str, qubit_indices: Tuple[int]):
        """
        Simplify the definition of Pauli Operator.

        Parameters
        ----------
        qubit_indices: `Tuple[int]`
                The qubit indices of the Pauli Operator.
        pauli_str: `str`
                The Pauli Operator basis string.

        Returns
        -------
        new_pauli_str: `str`
                The updated Pauli Operator basis string.
        new_qubit_indices: `Tuple[int]`
                The updated qubit indices in increasing order.

        Examples
        --------
        PauliOp('XZX',(3,2,2)) -> PauliOp('XY',(3,2)) with appropriate phase.
        """
        new_phase = 1

        qubit_reps = Counter(qubit_indices)

        # If no repetitions, do nothing
        if len(qubit_indices) == len(qubit_reps.values()):
            new_pauli_str = pauli_str
            new_qubit_indices = qubit_indices

        # If repetitions present, simplify operator
        else:
            # Extract spins with multiple operators
            repeating_indices = [index for index, rep in qubit_reps.items() if rep > 1]

            paulis_list_to_contract = []

            # Extract operators to be simpilfied
            for index in repeating_indices:
                paulis_list_to_contract.append(
                    [
                        pauli_str[i]
                        for i in range(len(qubit_indices))
                        if qubit_indices[i] == index
                    ]
                )

            # Simplify
            for paulis in paulis_list_to_contract:
                i = 0

                # Repeat until all operators have been simplified
                while len(paulis) > 1:
                    pauli_mult = paulis[i] + paulis[i + 1]
                    paulis[0] = PAULI_MULT_RULES[pauli_mult]
                    new_phase *= PAULI_PHASE_MAPPERS[pauli_mult]
                    paulis.pop(i + 1)

            # Store simplified strings and indices
            repeating_pauli_str = "".join(pauli[0] for pauli in paulis_list_to_contract)
            non_repeating_indices = [
                index for index, rep in qubit_reps.items() if rep == 1
            ]
            non_repeating_paulis = "".join(
                pauli_str[list(qubit_indices).index(idx)]
                for idx in non_repeating_indices
            )

            new_pauli_str = non_repeating_paulis + repeating_pauli_str
            new_qubit_indices = tuple(non_repeating_indices + repeating_indices)

        return new_pauli_str, new_qubit_indices, new_phase

    @property
    def _is_trivial(self) -> bool:
        """
        Returns `True` if the PauliOp only contains identity terms `'I'`.
        """
        return self.pauli_str == "I" * len(self.qubit_indices)

    @property
    def matrix(self):
        """
        Matrix representation of the Pauli Operator.
        """
        # Initialize matrix representation
        mat = PAULI_MAPPERS[self.pauli_str[0]]

        for pauli in self.pauli_str[1:]:
            mat = np.kron(mat, PAULI_MAPPERS[pauli])
        return mat

    def __len__(self):
        """
        Length of the Pauli term.
        """
        return len(self.qubit_indices)

    def __eq__(self, other_pauli_op):
        """
        Check whether two pauli_operators are equivalent, by comparing qubit indices and pauli strings.
        """
        tuple_1 = (self.qubit_indices, self.pauli_str, self.phase)
        tuple_2 = (
            other_pauli_op.qubit_indices,
            other_pauli_op.pauli_str,
            other_pauli_op.phase,
        )
        return tuple_1 == tuple_2

    def __copy__(self):
        """
        Create a new `PauliOp` by copying the current one.
        """
        copied_pauli_op = self.__class__.__new__(self.__class__)
        for attribute, value in vars(self).items():
            setattr(copied_pauli_op, attribute, value)
        return copied_pauli_op

    def __str__(self):
        """
        String representation of the Pauli Operator.
        """
        term_str = "".join(
            pauli_base + "_" + str({index})
            for pauli_base, index in zip(self.pauli_str, self.qubit_indices)
        )
        return term_str

    def __repr__(self):
        """
        String representation of the Pauli Operator object.
        """
        term_repr = f"PauliOp({self.pauli_str},{self.qubit_indices})"
        return term_repr

    def __mul__(self, other_pauli_op):
        """
        Multiply two Pauli Operators.

        Parameters
        ----------
        other_pauli_op: `PauliOp`
                The other Pauli Operator to be multiplied.

        Return
        ------
        new_pauli_op: `PauliOp`
                The resulting Pauli Operator after the multiplication.
        """
        assert isinstance(other_pauli_op, PauliOp), "Please specify a Pauli Operator"

        copied_current_pauli_op = self.__copy__()
        copied_current_pauli_op.__matmul__(other_pauli_op)

        return copied_current_pauli_op

    def __matmul__(self, other_pauli_op):
        """
        In-place Multiplication of Pauli Operators. Contract `other_pauli_op` into `self`.

        Parameters
        ----------
        other_pauli_op: `PauliOp`
                The Pauli Operator object to be multiplied.
        """
        assert isinstance(other_pauli_op, PauliOp), "Please specify a Pauli Operator"

        # Combined number of qubits
        n_qubits = max(max(self.qubit_indices), max(other_pauli_op.qubit_indices)) + 1

        self_pauli_str_list = list(self.pauli_str)
        other_pauli_str_list = list(other_pauli_op.pauli_str)

        # Fill pauli strings with identity operators for Pauli objects
        # for the number of strings to match the combined number of qubits
        for i in range(n_qubits):
            if i not in self.qubit_indices:
                self_pauli_str_list.insert(i, "I")
            if i not in other_pauli_op.qubit_indices:
                other_pauli_str_list.insert(i, "I")

        new_full_operator = ""
        new_phase = 1

        # Perform multiplication
        for idx in range(n_qubits):
            pauli_composition = self_pauli_str_list[idx] + other_pauli_str_list[idx]
            mult_phase = PAULI_PHASE_MAPPERS[pauli_composition]
            mult_pauli = PAULI_MULT_RULES[pauli_composition]

            new_full_operator += mult_pauli
            new_phase *= mult_phase

        self.qubit_indices = tuple(
            [idx for idx in range(n_qubits) if new_full_operator[idx] != "I"]
        )
        self.pauli_str = new_full_operator.replace("I", "")
        self.phase = new_phase

        return self

    @classmethod
    def X(cls, qubit_idx):
        """
        Pauli X operator.
        """
        return cls("X", (qubit_idx,))

    @classmethod
    def Y(cls, qubit_idx):
        """
        Pauli Y operator.
        """
        return cls("Y", (qubit_idx,))

    @classmethod
    def Z(cls, qubit_idx):
        """
        Pauli Z operator.
        """
        return cls("Z", (qubit_idx,))

    @classmethod
    def I(cls, qubit_idx):
        """
        Pauli identity operator.
        """
        return cls("I", (qubit_idx,))


class Hamiltonian:
    """
    General Hamiltonian class.

    Attributes
    ----------
    n_qubits: `int`
    terms: `List[PauliOp]`
    coeffs: `List[complex,float,int]`
    constant: `float`
    qubits_pairs: `List[PauliOp]`
    qubits_singles: `List[PauliOp]`
    single_qubit_coeffs: `List[float]`
    pair_qubit_coeffs: `List[float]`
    qureg
    expression
    hamiltonian_squared
    """

    def __init__(
        self,
        pauli_terms: List[PauliOp],
        coeffs: List[Union[complex, int, float]],
        constant: float,
        divide_into_singles_and_pairs: bool = True,
    ):
        """
        Parameters
        ----------
        pauli_terms: `List[PauliOp]`
            Set of terms in the Hamiltonian as PauliOp objects.
        coeffs: `List[Union[complex,float,int]]`
            Multiplicative coefficients for each Pauli term in the Hamiltonian.
        constant: `float`
            Constant term in the Hamiltonian.
        divide_into_singles_and_pairs: `bool`, optional
                Whether to divide the Hamiltonian into singles and pairs
        """
        assert len(pauli_terms) == len(
            coeffs
        ), "Number of Pauli terms in Hamiltonian should be same as number of coefficients"

        physical_qureg = []

        # Extract physical regiser from qubit indices
        for pauli_term in pauli_terms:
            if isinstance(pauli_term, PauliOp):
                physical_qureg.extend(pauli_term.qubit_indices)
            else:
                raise TypeError(
                    f"Pauli terms should be of type PauliOp and not {type(pauli_term)}"
                )
        physical_qureg = list(set(physical_qureg))

        # Number of qubits
        self.n_qubits = len(physical_qureg)

        # Extract qubit map if necessary
        need_remapping = False
        if physical_qureg != self.qureg:
            print(
                f"Qubits in the specified Hamiltonian are remapped to {self.qureg}."
                "Please specify the physical quantum register as a qubit layout argument in the backend"
            )
            need_remapping = True
            qubit_mapper = dict(zip(physical_qureg, self.qureg))

        self.terms = []
        self.coeffs = []
        self.constant = constant

        for term, coeff in zip(pauli_terms, coeffs):
            # Identity terms are added to the constant
            if term._is_trivial:
                self.constant += coeff

            # Remap terms if required
            else:
                if need_remapping:
                    new_indices = tuple(qubit_mapper[i] for i in term.qubit_indices)
                    pauli_str = term.pauli_str
                    self.terms.append(PauliOp(pauli_str, new_indices))
                else:
                    self.terms.append(term)

                # Update the coefficients with phase from Pauli Operators
                self.coeffs.append(coeff * term.phase)
                # after absorbing the phase in coeff, set the phase in term to 1
                term.phase = 1

        if divide_into_singles_and_pairs:
            self._divide_into_singles_pairs()

    @property
    def qureg(self):
        """
        List of qubits from 0 to n-1 in Hamiltonian.
        """
        return list(range(self.n_qubits))

    def _divide_into_singles_pairs(self):
        """
        Extract terms and coefficients for linear and quadratic terms in the Hamiltonian.
        """

        self.qubits_pairs = []
        self.qubits_singles = []
        self.single_qubit_coeffs = []
        self.pair_qubit_coeffs = []

        for term, coeff in zip(self.terms, self.coeffs):
            if len(term) == 1:
                self.qubits_singles.append(term)
                self.single_qubit_coeffs.append(coeff)
            elif len(term) == 2:
                self.qubits_pairs.append(term)
                self.pair_qubit_coeffs.append(coeff)
            else:
                raise NotImplementedError(
                    "Hamiltonian only supports Linear and Quadratic terms"
                )

    def __str__(self):
        """
        Return a string representation of the Hamiltonian.
        """
        hamil_str = ""
        for coeff, term in zip(self.coeffs, self.terms):
            hamil_str += str(np.round(coeff, 3)) + "*" + term.__str__() + " + "
        hamil_str += str(self.constant)
        return hamil_str

    def __len__(self):
        """
        Return the number of terms in the Hamiltonian.
        """
        return len(self.terms)

    @property
    def expression(self):
        """
        Generates a symbolic expression for the Hamiltonian.

        Returns
        -------
        hamiltonian_expression: `sympy.Symbol`
            Symbolic expression for the Hamiltonian.
        """

        # Generate expression
        hamiltonian_expression = Symbol(str(self.constant))
        for term, coeff in zip(self.terms, self.coeffs):
            hamiltonian_expression += Symbol(str(coeff) + term.__str__())
        return hamiltonian_expression

    # A function that outputs a dictionary of the Hamiltonian terms and coefficients
    def hamiltonian_dict(self, classical: bool = True):
        """
        Generates a dictionary of the Hamiltonian terms and coefficients.

        Parameters
        ----------
        classical: `bool`, optional
            If true, returns a dictionary containing only qubit indices as terms
            else returns a dictionary containing PauliOp objects as terms.

        Returns
        -------
        `dict`
            Dictionary of the Hamiltonian terms and coefficients.
        """
        hamiltonian_dict = {}
        for term, coeff in zip(self.terms, self.coeffs):
            if classical:
                hamiltonian_dict[tuple(term.qubit_indices)] = coeff
            else:
                hamiltonian_dict[term] = coeff
            # add the constant term
            hamiltonian_dict[()] = self.constant
        return hamiltonian_dict

    def __add__(self, other_hamiltonian):
        """
        Add two Hamiltonians in place updating `self`

        Parameters
        ----------
        other_hamiltonian: `Hamiltonian`
                The other Hamiltonian to be added
        """
        assert isinstance(other_hamiltonian, Hamiltonian)

        for other_term, other_coeff in zip(
            other_hamiltonian.terms, other_hamiltonian.coeffs
        ):
            if other_term in self.terms:
                self.coeffs[self.terms.index(other_term)] += other_coeff
            else:
                self.terms.append(other_term)
                self.coeffs.append(other_coeff)

    @property
    def hamiltonian_squared(self):
        """
        Compute the squared of the Hamiltonian, necessary for computing
        the error in expectation values.

        Returns
        -------
        hamil_squared: `Hamiltonian`
            Hamiltonian squared.
        """
        hamil_sq_terms = []
        hamil_sq_coeffs = []
        hamil_sq_constant = self.constant**2

        for i, term1 in enumerate(self.terms):
            for j, term2 in enumerate(self.terms):
                new_term = term1 * term2

                # If multiplication yields a constant add it to hamiltonian constant
                if new_term.pauli_str == "":
                    hamil_sq_constant += self.coeffs[i] * self.coeffs[j]

                # If it yields non-trivial term add it to list of terms
                else:
                    new_phase = new_term.phase
                    new_coeff = self.coeffs[i] * self.coeffs[j]
                    hamil_sq_terms.append(new_term)
                    hamil_sq_coeffs.append(new_coeff)

        hamil_sq_terms.extend(self.terms)
        hamil_sq_coeffs.extend([2 * self.constant * coeff for coeff in self.coeffs])

        hamil_squared = Hamiltonian(
            hamil_sq_terms,
            hamil_sq_coeffs,
            hamil_sq_constant,
            divide_into_singles_and_pairs=False,
        )
        return hamil_squared

    @classmethod
    def classical_hamiltonian(
        cls,
        terms: List[Union[Tuple, List]],
        coeffs: List[Union[float, int]],
        constant: float,
    ):
        """
        Generates a classical Hamiltonian from a list of terms, coefficients and constant.

        Parameters
        ----------
        terms: `List[tuple]` or `List[list]`
            Set of qubit indices for each term in the Hamiltonian.
        coeffs: `List[float]` or `List[int]`
            Coefficients associated with each term in the Hamiltonian
        constant: `float`
            Constant term in the Hamiltonian.

        Returns
        -------
        Hamiltonian:
            Classical Hamiltonian.
        """
        pauli_ops = []
        pauli_coeffs = []

        for term, coeff in zip(terms, coeffs):
            # Check coeffcient type
            if not isinstance(coeff, int) and not isinstance(coeff, float):
                raise ValueError(
                    "Classical Hamiltonians only support Integer or Float coefficients"
                )

            # Construct Hamiltonian terms from Pauli operators
            if len(term) == 2:
                pauli_ops.append(PauliOp("ZZ", term))
                pauli_coeffs.append(coeff)
            elif len(term) == 1:
                pauli_ops.append(PauliOp("Z", term))
                pauli_coeffs.append(coeff)
            elif len(term) == 0:
                constant += coeff
            else:
                raise ValueError("Hamiltonian only supports Linear and Quadratic terms")

        return cls(pauli_ops, pauli_coeffs, constant)

    @property
    def as_matrix(self):
        """
        Build sparse matrix of Hamiltonian. Note this includes self.constant value.

        Returns
        -------
        H_mat:
            sparse matrix of Hamiltonian.
        """

        if self.n_qubits > 20:
            raise ValueError("number of qubits exceeds maximum of 20")

        p_mat_dict = {
            "X": csr_matrix([[0, 1], [1, 0]]),
            "Y": csr_matrix([[0.0, -1j], [1j, 0.0]]),
            "Z": csr_matrix([[1, 0], [0, -1]]),
            "I": csr_matrix([[1, 0], [0, 1]]),
        }

        converter = lambda s: p_mat_dict[s]

        # vectorized function to map single qubit pauli strings to sparse matrix
        vfunc = np.vectorize(converter)

        base = np.array(["I" for _ in range(self.n_qubits)])

        H_mat = reduce(sparse_kron, vfunc(base)) * self.constant
        for P, coeff in zip(self.terms, self.coeffs):
            temp = base.copy()

            # replace single qubit pauli I in base term
            # with single qubit pauli operators according to qubit_indices and pauli_strs in P
            temp[np.array(P.qubit_indices)] = np.array(list(P.pauli_str))

            # take kronecker product of sparse list of matrices
            H_mat += reduce(sparse_kron, vfunc(temp)) * coeff

        return H_mat



import os
import numpy as np
import pickle
from copy import deepcopy
from abc import ABC, abstractmethod
from typing import Type, Callable, List
from datetime import datetime

from scipy.optimize._minimize import minimize, MINIMIZE_METHODS
from scipy.optimize import LinearConstraint, NonlinearConstraint, Bounds

from ..backends.basebackend import VQABaseBackend
from ..backends.qaoa_vectorized import QAOAvectorizedBackendSimulator
from ..qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from . import optimization_methods as om
from .pennylane import optimization_methods_pennylane as ompl

from .logger_vqa import Logger
from ..algorithms.qaoa.qaoa_result import QAOAResult

from ..derivatives.derivative_functions import derivative
from ..derivatives.qfim import qfim

###
# TODO: Find better place for this

import pandas as pd
from typing import Any


def save_parameter(parameter_name: str, parameter_value: Any):
    filename = "oq_saved_info_" + parameter_name

    try:
        opened_csv = pd.read_csv(filename + ".csv")
    except Exception:
        opened_csv = pd.DataFrame(columns=[parameter_name])

    if type(parameter_value) not in [str, float, int]:
        parameter_value = str(parameter_value)

    update_df = pd.DataFrame(data={parameter_name: parameter_value}, index=[0])
    new_df = pd.concat([opened_csv, update_df], ignore_index=True)

    new_df.to_csv(filename + ".csv", index=False)

    print("Parameter Saving Successful")


###


class OptimizeVQA(ABC):
    """
    Training Class for optimizing VQA algorithm that wraps around VQABaseBackend and QAOAVariationalBaseParams objects.
    This function utilizes the `update_from_raw` of the QAOAVariationalBaseParams class and `expectation` method of
    the VQABaseBackend class to create a wrapper callable which is passed into scipy.optimize.minimize for minimization.
    Only the trainable parameters should be passed instead of the complete
    AbstractParams object. The construction is completely backend and type
    of VQA agnostic.

    This class is an AbstractBaseClass on top of which other specific Optimizer
    classes are built.

    .. Tip::
        Optimizer that usually work the best for quantum optimization problems

        * Gradient free optimizer - Cobyla

    Parameters
    ----------
    vqa_object:
        Backend object of class VQABaseBackend which contains information on the
        backend used to perform computations, and the VQA circuit.
    variational_params:
        Object of class QAOAVariationalBaseParams, which contains information on the circuit to be executed,
        the type of parametrisation, and the angles of the VQA circuit.
    method:
        which method to use for optimization. Choose a method from the list
        of supported methods by scipy optimize, or from the list of custom gradient optimisers.
    optimizer_dict:
        All extra parameters needed for customising the optimising, as a dictionary.
    Optimizers that usually work the best for quantum optimization problems:
        #. Gradient free optimizer: BOBYQA, ImFil, Cobyla
        #. Gradient based optimizer: L-BFGS, ADAM (With parameter shift gradients)

        Note: Adam is not a part of scipy, it will added in a future version
    """

    def __init__(
        self,
        vqa_object: Type[VQABaseBackend],
        variational_params: Type[QAOAVariationalBaseParams],
        optimizer_dict: dict,
    ):
        if not isinstance(vqa_object, VQABaseBackend):
            raise TypeError(f"The specified cost object must be of type VQABaseBackend")

        self.vqa = vqa_object
        # extract initial parameters from the params of vqa_object
        self.variational_params = variational_params
        self.initial_params = variational_params.raw()
        self.method = optimizer_dict["method"].lower()
        self.save_to_csv = optimizer_dict.get("save_intermediate", False)

        self.log = Logger(
            {
                "cost": {
                    "history_update_bool": optimizer_dict.get("cost_progress", True),
                    "best_update_string": "LowestOnly",
                },
                "measurement_outcomes": {
                    "history_update_bool": optimizer_dict.get(
                        "optimization_progress", False
                    ),
                    "best_update_string": "Replace",
                },
                "param_log": {
                    "history_update_bool": optimizer_dict.get("parameter_log", True),
                    "best_update_string": "Replace",
                },
                "eval_number": {
                    "history_update_bool": False,
                    "best_update_string": "Replace",
                },
                "func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "jac_func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "qfim_func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "job_ids": {
                    "history_update_bool": True,
                    "best_update_string": "Replace",
                },
                "n_shots": {
                    "history_update_bool": True,
                    "best_update_string": "Replace",
                },
            },
            {
                "root_nodes": [
                    "cost",
                    "func_evals",
                    "jac_func_evals",
                    "qfim_func_evals",
                    "n_shots",
                ],
                "best_update_structure": (
                    ["cost", "param_log"],
                    ["cost", "measurement_outcomes"],
                    ["cost", "job_ids"],
                    ["cost", "eval_number"],
                ),
            },
        )

        self.log.log_variables(
            {"func_evals": 0, "jac_func_evals": 0, "qfim_func_evals": 0}
        )

    @abstractmethod
    def __repr__(self):
        """
        Overview of the instantiated optimier/trainer.
        """
        string = f"Optimizer for VQA of type: {type(self.vqa).__base__.__name__} \n"
        string += f"Backend: {type(self.vqa).__name__} \n"
        string += f"Method: {str(self.method).upper()}\n"

        return string

    def __call__(self):
        """
        Call the class instance to initiate the training process.
        """
        self.optimize()
        return self

    # def evaluate_jac(self, x):

    def optimize_this(self, x, n_shots=None):
        """
        A function wrapper to execute the circuit in the backend. This function
        will be passed as argument to be optimized by scipy optimize.

        Parameters
        ----------
        x:
            Parameters (a list of floats) over which optimization is performed.
        n_shots:
            Number of shots to be used for the backend when computing the expectation.
            If None, nothing is passed to the backend.

        Returns
        -------
        cost value:
            Cost value which is evaluated on the declared backend.

        Returns
        -------
        :
            Cost Value evaluated on the declared backed or on the Wavefunction Simulator if specified so
        """

        log_dict = {}

        self.variational_params.update_from_raw(deepcopy(x))
        log_dict.update({"param_log": self.variational_params.raw()})

        if hasattr(self.vqa, "log_with_backend") and callable(
            getattr(self.vqa, "log_with_backend")
        ):
            self.vqa.log_with_backend(
                metric_name="variational_params",
                value=self.variational_params,
                iteration_number=self.log.func_evals.best[0],
            )

        if self.save_to_csv:
            save_parameter("param_log", deepcopy(x))

        n_shots_dict = {"n_shots": n_shots} if n_shots else {}
        callback_cost = self.vqa.expectation(self.variational_params, **n_shots_dict)

        log_dict.update({"cost": callback_cost})

        current_eval = self.log.func_evals.best[0]
        current_eval += 1
        log_dict.update({"func_evals": current_eval})

        log_dict.update(
            {
                "eval_number": (
                    current_eval
                    - self.log.jac_func_evals.best[0]
                    - self.log.qfim_func_evals.best[0]
                )
            }
        )  # this one will say which evaluation is the optimized one

        log_dict.update({"measurement_outcomes": self.vqa.measurement_outcomes})

        if hasattr(self.vqa, "log_with_backend") and callable(
            getattr(self.vqa, "log_with_backend")
        ):
            self.vqa.log_with_backend(
                metric_name="measurement_outcomes",
                value=self.vqa.measurement_outcomes,
                iteration_number=self.log.func_evals.best[0],
            )

        if hasattr(self.vqa, "job_id"):
            log_dict.update({"job_ids": self.vqa.job_id})

            if self.save_to_csv:
                save_parameter("job_ids", self.vqa.job_id)

        self.log.log_variables(log_dict)

        return callback_cost

    @abstractmethod
    def optimize(self):
        """
        Main method which implements the optimization process.
        Child classes must implement this method according to their respective
        optimization process.

        Returns
        -------
        :
            The optimized return object from the ``scipy.optimize`` package the result
            is assigned to the attribute ``opt_result``
        """
        pass

    def results_dictionary(self, file_path: str = None, file_name: str = None):
        """
        This method formats a dictionary that consists of all the results from
        the optimization process. The dictionary is returned by this method.
        The results can also be saved by providing the path to save the pickled file

        .. Important::
            Child classes must implement this method so that the returned object,
            a ``Dictionary`` is consistent across all Optimizers.

        TODO:
            Decide results datatype: dictionary or namedtuple?

        Parameters
        ----------
        file_path:
            To save the results locally on the machine in pickle format, specify
            the entire file path to save the result_dictionary.

        file_name:
            Custom name for to save the data; a generic name with the time of
            optimization is used if not specified
        """
        date_time = datetime.now().strftime("%d.%m.%Y_%H.%M.%S")
        file_name = f"opt_results_{date_time}" if file_name is None else file_name

        self.qaoa_result = QAOAResult(
            self.log, self.method, self.vqa.cost_hamiltonian, type(self.vqa)
        )

        if file_path and os.path.isdir(file_path):
            print("Saving results locally")
            pickled_file = open(f"{file_path}/{file_name}.pcl", "wb")
            pickle.dump(self.qaoa_result, pickled_file)
            pickled_file.close()

        return  # result_dict


class ScipyOptimizer(OptimizeVQA):
    """
    Python vanilla scipy based optimizer for the VQA class.

    .. Tip::
        Using bounds may result in lower optimization performance

    Parameters
    ----------
    vqa_object:
        Backend object of class VQABaseBackend which contains information on the
        backend used to perform computations, and the VQA circuit.

    variational_params:
        Object of class QAOAVariationalBaseParams, which contains information on
        the circuit to be executed,  the type of parametrisation,
        and the angles of the VQA circuit.

    optimizer_dict:
        * 'jac': gradient as ``Callable``, if defined else ``None``
        * 'hess': hessian as ``Callable``, if defined else ``None``
        * 'bounds': parameter bounds while training, defaults to ``None``
        * 'constraints': Linear/Non-Linear constraints (only for COBYLA, SLSQP and trust-constr)
        * 'tol': Tolerance for termination
        * 'maxiter': sets ``maxiters = 100`` by default if not specified.
        * 'maxfev': sets ``maxfev = 100`` by default if not specified.
        * 'optimizer_options': dictionary of optimiser-specific arguments, defaults to ``None``
    """

    GRADIENT_FREE = ["cobyla", "nelder-mead", "powell", "slsqp"]
    SCIPY_METHODS = MINIMIZE_METHODS

    def __init__(
        self,
        vqa_object: Type[VQABaseBackend],
        variational_params: Type[QAOAVariationalBaseParams],
        optimizer_dict: dict,
    ):
        super().__init__(vqa_object, variational_params, optimizer_dict)

        self.vqa_object = vqa_object
        self._validate_and_set_params(optimizer_dict)

    def _validate_and_set_params(self, optimizer_dict):
        """
        Verify that the specified arguments are valid for the particular optimizer.
        """

        if self.method not in ScipyOptimizer.SCIPY_METHODS:
            raise ValueError("Specified method not supported by Scipy Minimize")

        jac = optimizer_dict.get("jac", None)
        hess = optimizer_dict.get("hess", None)
        jac_options = optimizer_dict.get("jac_options", None)
        hess_options = optimizer_dict.get("hess_options", None)

        if self.method not in ScipyOptimizer.GRADIENT_FREE and (
            jac is None or not isinstance(jac, (Callable, str))
        ):
            raise ValueError(
                "Please specify either a string or provide callable"
                "gradient in order to use gradient based methods"
            )
        else:
            if isinstance(jac, str):
                self.jac = derivative(
                    self.vqa_object,
                    self.variational_params,
                    self.log,
                    "gradient",
                    jac,
                    jac_options,
                )
            else:
                self.jac = jac

        hess = optimizer_dict.get("hess", None)
        if hess is not None and not isinstance(hess, (Callable, str)):
            raise ValueError("Hessian needs to be of type Callable or str")
        else:
            if isinstance(hess, str):
                self.hess = derivative(
                    self.vqa_object,
                    self.variational_params,
                    self.log,
                    "hessian",
                    hess,
                    hess_options,
                )
            else:
                self.hess = hess

        constraints = optimizer_dict.get("constraints", ())
        if (
            constraints == ()
            or isinstance(constraints, LinearConstraint)
            or isinstance(constraints, NonlinearConstraint)
        ):
            self.constraints = constraints
        else:
            raise ValueError(
                f"Constraints for Scipy optimization should be of type {LinearConstraint} or {NonlinearConstraint}"
            )

        bounds = optimizer_dict.get("bounds", None)

        if bounds is None or isinstance(bounds, Bounds):
            self.bounds = bounds
        elif isinstance(bounds, List):
            lb = np.array(bounds).T[0]
            ub = np.array(bounds).T[1]
            self.bounds = Bounds(lb, ub)
        else:
            raise ValueError(
                f"Bounds for Scipy optimization should be of type {Bounds},"
                "or a list in the form [[ub1, lb1], [ub2, lb2], ...]"
            )

        self.options = optimizer_dict.get("optimizer_options", {})
        self.options["maxiter"] = optimizer_dict.get("maxiter", None)
        if optimizer_dict.get("maxfev") is not None:
            self.options["maxfev"] = optimizer_dict.get("maxfev", None)

        self.tol = optimizer_dict.get("tol", None)

        return self

    def __repr__(self):
        """
        Overview of the instantiated optimier/trainer.
        """
        maxiter = self.options["maxiter"]
        string = f"Optimizer for VQA of type: {type(self.vqa).__base__.__name__} \n"
        string += f"Backend: {type(self.vqa).__name__} \n"
        string += f"Method: {str(self.method).upper()} with Max Iterations: {maxiter}\n"

        return string

    def optimize(self):
        """
        Main method which implements the optimization process using ``scipy.minimize``.

        Returns
        -------
        :
            Returns self after the optimization process is completed.
        """

        try:
            if self.method not in ScipyOptimizer.GRADIENT_FREE:
                if self.hess == None:
                    result = minimize(
                        self.optimize_this,
                        x0=self.initial_params,
                        method=self.method,
                        jac=self.jac,
                        tol=self.tol,
                        constraints=self.constraints,
                        options=self.options,
                        bounds=self.bounds,
                    )
                else:
                    result = minimize(
                        self.optimize_this,
                        x0=self.initial_params,
                        method=self.method,
                        jac=self.jac,
                        hess=self.hess,
                        tol=self.tol,
                        constraints=self.constraints,
                        options=self.options,
                        bounds=self.bounds,
                    )
            else:
                result = minimize(
                    self.optimize_this,
                    x0=self.initial_params,
                    method=self.method,
                    tol=self.tol,
                    constraints=self.constraints,
                    options=self.options,
                    bounds=self.bounds,
                )
        except ConnectionError as e:
            print(e, "\n")
            print(
                "The optimization has been terminated early. Most likely due to a connection error."
                "You can retrieve results from the optimization runs that were completed"
                "through the .result method."
            )
        except Exception as e:
            raise e
        finally:
            self.results_dictionary()

        return self


class CustomScipyGradientOptimizer(OptimizeVQA):
    """
    Python custom scipy gradient based optimizer for the VQA class.

    .. Tip::
        Using bounds may result in lower optimization performance

    Parameters
    ----------
    vqa_object:
        Backend object of class VQABaseBackend which contains information on
        the backend used to perform computations, and the VQA circuit.

    variational_params:
        Object of class QAOAVariationalBaseParams, which contains information on
        the circuit to be executed,  the type of parametrisation,
        and the angles of the VQA circuit.

    optimizer_dict:
        * 'jac': gradient as ``Callable``, if defined else ``None``
        * 'hess': hessian as ``Callable``, if defined else ``None``
        * 'bounds': parameter bounds while training, defaults to ``None``
        * 'constraints': Linear/Non-Linear constraints (only for COBYLA, SLSQP and trust-constr)
        * 'tol': Tolerance for termination
        * 'maxiter': sets ``maxiters = 100`` by default if not specified.
        * 'maxfev': sets ``maxfev = 100`` by default if not specified.
        * 'optimizer_options': dictionary of optimiser-specific arguments, defaults to ``None``

    """

    CUSTOM_GRADIENT_OPTIMIZERS_MAPPER = {
        "vgd": om.grad_descent,
        "newton": om.newton_descent,
        "rmsprop": om.rmsprop,
        "natural_grad_descent": om.natural_grad_descent,
        "spsa": om.SPSA,
        "cans": om.CANS,
        "icans": om.iCANS,
    }
    CUSTOM_GRADIENT_OPTIMIZERS = list(CUSTOM_GRADIENT_OPTIMIZERS_MAPPER.keys())

    def __init__(
        self,
        vqa_object: Type[VQABaseBackend],
        variational_params: Type[QAOAVariationalBaseParams],
        optimizer_dict: dict,
    ):
        super().__init__(vqa_object, variational_params, optimizer_dict)

        self.vqa_object = vqa_object
        self._validate_and_set_params(optimizer_dict)

    def _validate_and_set_params(self, optimizer_dict):
        """
        Verify that the specified arguments are valid for the particular optimizer.
        """

        if self.method not in CustomScipyGradientOptimizer.CUSTOM_GRADIENT_OPTIMIZERS:
            raise ValueError(
                f"Please choose from the supported methods: {CustomScipyGradientOptimizer.CUSTOM_GRADIENT_OPTIMIZERS}"
            )

        jac = optimizer_dict.get("jac", None)
        hess = optimizer_dict.get("hess", None)
        jac_options = optimizer_dict.get("jac_options", None)
        hess_options = optimizer_dict.get("hess_options", None)

        if jac is None or not isinstance(jac, (Callable, str)):
            raise ValueError(
                "Please specify either a string or provide callable gradient in order to use gradient based methods"
            )
        else:
            if isinstance(jac, str):
                if not self.method in ["icans", "cans"]:
                    self.jac = derivative(
                        self.vqa_object,
                        self.variational_params,
                        self.log,
                        "gradient",
                        jac,
                        jac_options,
                    )
                else:
                    self.jac = None
                    self.jac_w_variance = derivative(
                        self.vqa_object,
                        self.variational_params,
                        self.log,
                        "gradient_w_variance",
                        jac,
                        jac_options,
                    )

            else:
                self.jac = jac

        if hess is not None and not isinstance(hess, (Callable, str)):
            raise ValueError("Hessian needs to be of type Callable or str")
        else:
            if isinstance(hess, str):
                self.hess = derivative(
                    self.vqa_object,
                    self.variational_params,
                    self.log,
                    "hessian",
                    hess,
                    hess_options,
                )
            else:
                self.hess = hess

        constraints = optimizer_dict.get("constraints", ())
        if (
            constraints == ()
            or isinstance(constraints, LinearConstraint)
            or isinstance(constraints, NonlinearConstraint)
        ):
            self.constraints = constraints
        else:
            raise ValueError(
                f"Constraints for Scipy optimization should be of type {LinearConstraint} or {NonlinearConstraint}"
            )

        bounds = optimizer_dict.get("bounds", None)
        if bounds is None or isinstance(bounds, Bounds):
            self.bounds = bounds
        else:
            raise ValueError(
                f"Bounds for Scipy optimization should be of type {Bounds}"
            )

        self.options = optimizer_dict.get("optimizer_options", {})
        self.options["maxiter"] = optimizer_dict.get("maxiter", None)
        if optimizer_dict.get("maxfev") is not None:
            self.options["maxfev"] = optimizer_dict.get("maxfev", None)

        self.tol = optimizer_dict.get("tol", None)

        return self

    def __repr__(self):
        """
        Overview of the instantiated optimier/trainer.
        """
        maxiter = self.options["maxiter"]
        string = f"Optimizer for VQA of type: {type(self.vqa).__base__.__name__} \n"
        string += f"Backend: {type(self.vqa).__name__} \n"
        string += f"Method: {str(self.method).upper()} with Max Iterations: {maxiter}\n"

        return string

    def optimize(self):
        """
        Main method which implements the optimization process using ``scipy.minimize``.

        Returns
        -------
        :
            Returns self after the optimization process is completed.
            The optimized result is assigned to the attribute ``opt_result``
        """

        # set the optimizer function to the `method` variable based on the input

        method = CustomScipyGradientOptimizer.CUSTOM_GRADIENT_OPTIMIZERS_MAPPER[
            self.method
        ]

        if self.method == "natural_grad_descent":
            self.options["qfim"] = qfim(
                self.vqa_object, self.variational_params, self.log
            )
        elif self.method == "spsa":
            print("Warning : SPSA is an experimental feature.")
        elif self.method in ["cans", "icans"]:
            self.options["jac_w_variance"] = self.jac_w_variance
            self.options["coeffs"] = self.vqa_object.cost_hamiltonian.coeffs
            self.options["n_shots_cost"] = (
                self.vqa_object.n_shots
                if not isinstance(self.vqa_object, QAOAvectorizedBackendSimulator)
                else 0
            )

        try:
            if self.hess == None:
                result = minimize(
                    self.optimize_this,
                    x0=self.initial_params,
                    method=method,
                    jac=self.jac,
                    tol=self.tol,
                    constraints=self.constraints,
                    options=self.options,
                    bounds=self.bounds,
                )
            else:
                result = minimize(
                    self.optimize_this,
                    x0=self.initial_params,
                    method=method,
                    jac=self.jac,
                    hess=self.hess,
                    tol=self.tol,
                    constraints=self.constraints,
                    options=self.options,
                    bounds=self.bounds,
                )
        except ConnectionError as e:
            print(e, "\n")
            print(
                "The optimization has been terminated early. Most likely due to a connection error."
                "You can retrieve results from the optimization runs that were completed"
                "through the .result method."
            )
        except Exception as e:
            raise e
        finally:
            self.results_dictionary()

        return self


class PennyLaneOptimizer(OptimizeVQA):
    """
    Python custom scipy optimization with pennylane optimizers for the VQA class.

    .. Tip::
        Using bounds may result in lower optimization performance

    Parameters
    ----------
    vqa_object:
        Backend object of class VQABaseBackend which contains information on
        the backend used to perform computations, and the VQA circuit.

    variational_params:
        Object of class QAOAVariationalBaseParams, which contains information on
        the circuit to be executed,  the type of parametrisation,
        and the angles of the VQA circuit.

    optimizer_dict:
        * 'jac': gradient as ``Callable``, if defined else ``None``
        * 'hess': hessian as ``Callable``, if defined else ``None``
        * 'bounds': parameter bounds while training, defaults to ``None``
        * 'constraints': Linear/Non-Linear constraints (only for COBYLA, SLSQP and trust-constr)
        * 'tol': Tolerance for termination
        * 'maxiter': sets ``maxiters = 100`` by default if not specified.
        * 'maxfev': sets ``maxfev = 100`` by default if not specified.
        * 'optimizer_options': dictionary of optimiser-specific arguments, defaults to ``None``. Used also for the pennylande optimizers (and step function) arguments.

    """

    PENNYLANE_OPTIMIZERS = list(ompl.AVAILABLE_OPTIMIZERS.keys())

    def __init__(
        self,
        vqa_object: Type[VQABaseBackend],
        variational_params: Type[QAOAVariationalBaseParams],
        optimizer_dict: dict,
    ):
        super().__init__(vqa_object, variational_params, optimizer_dict)

        self.vqa_object = vqa_object
        self._validate_and_set_params(optimizer_dict)

    def _validate_and_set_params(self, optimizer_dict):
        """
        Verify that the specified arguments are valid for the particular optimizer.
        """

        if self.method not in PennyLaneOptimizer.PENNYLANE_OPTIMIZERS:
            raise ValueError(
                f"Please choose from the supported methods: {PennyLaneOptimizer.PENNYLANE_OPTIMIZERS}"
            )

        jac = optimizer_dict.get("jac", None)
        jac_options = optimizer_dict.get("jac_options", None)

        if jac is None or not isinstance(jac, (Callable, str)):
            raise ValueError(
                "Please specify either a string or provide callable gradient in order to use gradient based methods"
            )
        else:
            if isinstance(jac, str):
                self.jac = derivative(
                    self.vqa_object,
                    self.variational_params,
                    self.log,
                    "gradient",
                    jac,
                    jac_options,
                )
            else:
                self.jac = jac

        constraints = optimizer_dict.get("constraints", ())
        if (
            constraints == ()
            or isinstance(constraints, LinearConstraint)
            or isinstance(constraints, NonlinearConstraint)
        ):
            self.constraints = constraints
        else:
            raise ValueError(
                f"Constraints for Scipy optimization should be of type {LinearConstraint} or {NonlinearConstraint}"
            )

        bounds = optimizer_dict.get("bounds", None)
        if bounds is None or isinstance(bounds, Bounds):
            self.bounds = bounds
        else:
            raise ValueError(
                f"Bounds for Scipy optimization should be of type {Bounds}"
            )

        self.options = optimizer_dict.get("optimizer_options", {})
        self.options["maxiter"] = optimizer_dict.get("maxiter", None)
        if optimizer_dict.get("maxfev") is not None:
            self.options["maxfev"] = optimizer_dict.get("maxfev", None)

        self.tol = optimizer_dict.get("tol", None)

        return self

    def __repr__(self):
        """
        Overview of the instantiated optimier/trainer.
        """
        maxiter = self.options["maxiter"]
        string = f"Optimizer for VQA of type: {type(self.vqa).__base__.__name__} \n"
        string += f"Backend: {type(self.vqa).__name__} \n"
        string += f"Method: {str(self.method).upper()} with Max Iterations: {maxiter}\n"

        return string

    def optimize(self):
        """
        Main method which implements the optimization process using ``scipy.minimize``.

        Returns
        -------
        :
            Returns self after the optimization process is completed. The optimized result is assigned to the attribute ``opt_result``
        """

        # set the optimizer function
        method = ompl.pennylane_optimizer

        # set the method to be used for the optimization
        self.options["pennylane_method"] = self.method

        if self.options["pennylane_method"] in [
            "pennylane_spsa",
            "pennylane_rotosolve",
        ]:
            self.jac = None

        try:
            result = minimize(
                self.optimize_this,
                x0=self.initial_params,
                method=method,
                jac=self.jac,
                tol=self.tol,
                constraints=self.constraints,
                options=self.options,
                bounds=self.bounds,
            )
        except ConnectionError as e:
            print(e, "\n")
            print(
                "The optimization has been terminated early. Most likely due to a connection error."
                "You can retrieve results from the optimization runs that were completed"
                "through the .result method."
            )
        except Exception as e:
            raise e
        finally:
            self.results_dictionary()

        return self



from typing import Callable
from ..qaoa_components.ansatz_constructor import gates as gates_core
from ..qaoa_components.ansatz_constructor.rotationangle import RotationAngle


class VectorizedGateApplicator(gates_core.GateApplicator):
    VECTORIZED_OQ_GATE_MAPPER = lambda x: {
        gates_core.X.__name__: x.apply_x,
        gates_core.RZ.__name__: x.apply_rz,
        gates_core.RX.__name__: x.apply_rx,
        gates_core.RY.__name__: x.apply_ry,
        gates_core.RXX.__name__: x.apply_rxx,
        gates_core.RZX.__name__: x.apply_rzx,
        gates_core.RZZ.__name__: x.apply_rzz,
        gates_core.RYY.__name__: x.apply_ryy,
        gates_core.RYZ.__name__: x.apply_ryz,
        gates_core.RiSWAP.__name__: x.apply_rxy,
    }

    library = "vectorized"

    def gate_selector(
        self,
        gate: gates_core.Gate,
        vectorized_backend: "QAOAvectorizedBackendSimulator",
    ):
        selected_gate = VectorizedGateApplicator.VECTORIZED_OQ_GATE_MAPPER(
            vectorized_backend
        )[gate.__name__]
        return selected_gate

    @staticmethod
    def apply_1q_rotation_gate(
        vectorized_gate: Callable, qubit_1: int, rotation_object: RotationAngle
    ):
        vectorized_gate(qubit_1, rotation_object.rotation_angle)

    @staticmethod
    def apply_2q_rotation_gate(
        vectorized_gate: Callable,
        qubit_1: int,
        qubit_2: int,
        rotation_object: RotationAngle,
    ):
        vectorized_gate(qubit_1, qubit_2, rotation_object.rotation_angle)

    @staticmethod
    def apply_1q_fixed_gate(vectorized_gate: Callable, qubit_1: int):
        vectorized_gate(qubit_1)

    @staticmethod
    def apply_2q_fixed_gate(vectorized_gate: Callable, qubit_1: int, qubit_2: int):
        vectorized_gate(qubit_1, qubit_2)

    def apply_gate(self, gate: gates_core.Gate, *args):
        selected_vector_gate = self.gate_selector(gate, args[-1])
        # Remove argument from tuple
        args = args[:-1]
        if gate.n_qubits == 1:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,rotation_object)
                self.apply_1q_rotation_gate(selected_vector_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1)
                self.apply_1q_fixed_gate(selected_vector_gate, *args)
        elif gate.n_qubits == 2:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,qubit_2,rotation_object)
                self.apply_2q_rotation_gate(selected_vector_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1,qubit_2)
                self.apply_2q_fixed_gate(selected_vector_gate, *args)
        else:
            raise ValueError(
                "Error applying the requested gate. Please check in the input"
            )



from __future__ import annotations
from typing import List, Union, Callable


class RotationAngle(object):
    def __init__(
        self,
        angle_relationship: Callable,
        gate_label: GateMapLabel,
        value: Union[int, float] = None,
    ):
        """
        Angle object as placeholder for assigning angles to the parameterized
        gates in the circuit

        Parameters
        ----------
        angle_relationship: `Callable`
            A function that takes input a parameter and assigns
            the angle to the gate depending on the relationship
            between the parameter and the angle
        gate_label: `GateMapLabel`
            The label for the gatemap object to which the rotationangle
            is assigned
        value: `int` or `float`
            Value of the parameter
        """

        self._angle = angle_relationship
        self.gate_label = gate_label
        self.value = value

    @property
    def rotation_angle(self) -> Union[int, float]:
        return self._angle(self.value)



import itertools
import numpy as np

from ..utilities import check_kwargs
from .problem import Problem
from .qubo import QUBO


class Knapsack(Problem):
    """
    Creates an instance of the Kanpsack problem.

    Parameters
    ----------
    values: List[int]
        The values of the items that can be placed in the kanpsack.
    weights: List[int]
        The weight of the items that can be placed in the knapsack.
    weight_capacity: int
        The maximum weight the knapsack can hold.
    penalty: float
        Penalty for the weight constraint.

    Returns
    -------
        An instance of the Knapsack problem.
    """

    __name__ = "knapsack"

    def __init__(self, values, weights, weight_capacity, penalty):
        # Check whether the input is valid. Number of values should match the number of weights.
        if len(values) != len(weights):
            raise ValueError("Number of items does not match given value and weights")

        self.values = values
        self.weights = weights
        self.weight_capacity = weight_capacity
        self.penalty = penalty
        self.n_items = len(weights)

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, input_values):
        if not isinstance(input_values, list):
            raise TypeError("The input parameter, values, has to be a list")

        for each_entry in input_values:
            if not isinstance(each_entry, int):
                raise TypeError("The elements in values list must be of type int.")

        self._values = input_values

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, input_weights):
        if not isinstance(input_weights, list):
            raise TypeError("The input parameter, weights, has to be a list")

        for each_entry in input_weights:
            if not isinstance(each_entry, int):
                raise TypeError("The elements in weights list must be of type int.")

        self._weights = input_weights

    @property
    def weight_capacity(self):
        return self._weight_capacity

    @weight_capacity.setter
    def weight_capacity(self, input_weight_capacity):
        if not isinstance(input_weight_capacity, int):
            raise TypeError(
                "The input parameter, weight_capacity, has to be of type int"
            )

        if input_weight_capacity <= 0:
            raise TypeError(
                "The input parameter, weight_capacity, must be a positive integer greater than 0"
            )

        self._weight_capacity = input_weight_capacity

    @property
    def penalty(self):
        return self._penalty

    @penalty.setter
    def penalty(self, input_penalty):
        if not isinstance(input_penalty, int) and not isinstance(input_penalty, float):
            raise TypeError(
                "The input parameter, penalty, has to be of type float or int"
            )

        self._penalty = input_penalty

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Knapsack problem.

        Parameters
        ----------
        n_items: int
            The number of items that can be placed in the knapsack.

        Returns
        -------
            A random instance of the Knapsack problem.
        """
        n_items = check_kwargs(["n_items"], [None], **kwargs)[0]

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))

        min_weights = np.min(weights)
        max_weights = np.max(weights)

        if min_weights != max_weights:
            weight_capacity = int(
                rng.integers(min_weights * n_items, max_weights * n_items)
            )
        else:
            weight_capacity = int(rng.integers(max_weights, max_weights * n_items))

        penalty = 2 * np.max(values)
        return Knapsack(values, weights, weight_capacity, int(penalty))

    def terms_and_weights(self):
        n_variables_slack = int(np.ceil(np.log2(self.weight_capacity)))
        n_variables = self.n_items + n_variables_slack

        # Edges between variables to represent slack value (the s_j's)
        edges_slacks = itertools.combinations(range(n_variables_slack), 2)
        edges_slacks_with_weights = [
            (list(e), 2 * self.penalty * (2 ** e[0]) * (2 ** e[1]))
            for e in edges_slacks
        ]

        # Edges between decision variables for weights (the x_i's)
        edges_decision_vars = itertools.combinations(
            range(n_variables_slack, n_variables), 2
        )
        edges_decision_vars_with_weights = [
            (
                list(e),
                2
                * self.penalty
                * self.weights[e[0] - n_variables_slack]
                * self.weights[e[1] - n_variables_slack],
            )
            for e in edges_decision_vars
        ]

        # Edges between decisions and variables to represent slack value (the x_i's and s_j's)
        edges_slacks_decision_vars = itertools.product(
            range(n_variables_slack), range(n_variables_slack, n_variables)
        )
        edges_slacks_decision_vars_with_weights = [
            (
                list(e),
                2 * self.penalty * (2 ** e[0]) * self.weights[e[1] - n_variables_slack],
            )
            for e in edges_slacks_decision_vars
        ]

        # Linear terms for the variables to represent slack value (s_j's)
        single_interaction_slacks = [
            ([i], self.penalty * (2 ** (2 * i) - 2 * self.weight_capacity * 2**i))
            for i in range(n_variables_slack)
        ]

        # Linear terms for the decision variables (the x_i's)
        single_interaction_decisions_vars = [
            (
                [i],
                self.penalty * self.weights[i - n_variables_slack] ** 2
                - 2
                * self.penalty
                * self.weight_capacity
                * self.weights[i - n_variables_slack]
                - self.values[i - n_variables_slack],
            )
            for i in range(n_variables_slack, n_variables)
        ]

        # The constant term
        constant_term = [([], self.penalty * self.weight_capacity**2)]

        # Unzip to retrieve terms and weights in separate sequences
        return tuple(
            zip(
                *(
                    edges_slacks_with_weights
                    + edges_decision_vars_with_weights
                    + edges_slacks_decision_vars_with_weights
                    + single_interaction_slacks
                    + single_interaction_decisions_vars
                    + constant_term
                )
            )
        )

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        n_variables_slack = int(np.ceil(np.log2(self.weight_capacity)))
        n = self.n_items + n_variables_slack
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(n, terms, weights)
        return QUBO(n, ising_terms, ising_weights, self.problem_instance)


class SlackFreeKnapsack(Knapsack):
    """
    A slack variable free approach to the Knapsack problem Hamiltonian.
    The Hamiltonian consists of decision qubits with a quadratic penalty term centred
    on `W`, i.e. the maximum Knapsack Capacity.
    Creates an instance of the SlackFreeKanpsack problem.

    Parameters
    ----------
    values: List[int]
        The values of the items that can be placed in the kanpsack.
    weights: List[int]
        The weight of the items that can be placed in the knapsack.
    weight_capacity: int
        The maximum weight the knapsack can hold.
    penalty: float
        Penalty for the weight constraint.

    Returns
    -------
        An instance of the SlackFreeKnapsack problem.
    """

    __name__ = "slack_free_knapsack"

    def __init__(self, values, weights, weight_capacity, penalty):
        super().__init__(values, weights, weight_capacity, penalty)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Knapsack problem.

        Parameters
        ----------
        n_items: int
            The number of items that can be placed in the knapsack.

        Returns
        -------
            A random instance of the Knapsack problem.
        """
        n_items = check_kwargs(["n_items"], [None], **kwargs)[0]

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))

        min_weights = np.min(weights)
        max_weights = np.max(weights)
        if min_weights != max_weights:
            weight_capacity = int(
                rng.integers(min_weights * n_items, max_weights * n_items)
            )
        else:
            weight_capacity = int(rng.integers(max_weights, max_weights * n_items))

        penalty = 2 * np.max(values)
        return SlackFreeKnapsack(values, weights, weight_capacity, int(penalty))

    def terms_and_weights(self):
        """
        Implementation of single and two-qubit terms in the slack-free Hamiltonian
        for the Knapsack problem.
        """

        n_variables = self.n_items

        # Edges between decision variables for weights (the x_i's)
        edges_decision_vars = itertools.combinations(range(n_variables), 2)
        edges_decision_vars_with_weights = [
            (list(e), 2 * self.penalty * self.weights[e[0]] * self.weights[e[1]])
            for e in edges_decision_vars
        ]

        # Linear terms for the decision variables (the x_i's)
        single_interaction_decisions_vars = [
            (
                [i],
                self.penalty * self.weights[i] ** 2
                - 2 * self.penalty * self.weight_capacity * self.weights[i]
                - self.values[i],
            )
            for i in range(n_variables)
        ]

        # The constant term
        constant_term = [([], self.penalty * self.weight_capacity**2)]

        # Unzip to retrieve terms and weights in separate sequences
        return tuple(
            zip(
                *(
                    edges_decision_vars_with_weights
                    + single_interaction_decisions_vars
                    + constant_term
                )
            )
        )

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        n = self.n_items
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(n, terms, weights)
        return QUBO(n, ising_terms, ising_weights, self.problem_instance)



import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import MIS


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool


class TestMIS(unittest.TestCase):
    """Tests for the MIS class"""

    def test_mis_terms_weights_constant(self):
        """Test that MIS creates a correct QUBO from the provided graph"""

        gr = nx.generators.random_graphs.fast_gnp_random_graph(n=5, p=0.8, seed=1234)
        gr_edges = [
            [0, 3],
            [0, 4],
            [1, 2],
            [1, 3],
            [2, 4],
            [3, 4],
            [0],
            [1],
            [2],
            [3],
            [4],
        ]
        gr_weights = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, -0.5, -0.5, -0.5, -1.0, -1.0]

        mis_prob_qubo = MIS(gr).qubo

        self.assertTrue(terms_list_equality(gr_edges, mis_prob_qubo.terms))
        self.assertEqual(gr_weights, mis_prob_qubo.weights)
        self.assertEqual(0.5, mis_prob_qubo.constant)

    def test_mis_random_problem(self):
        """Test MIS random instance method"""

        seed = 1234
        gr = nx.generators.random_graphs.fast_gnp_random_graph(n=10, p=0.8, seed=seed)
        mis_manual_prob = MIS(gr).qubo

        np.random.seed(1234)
        mis_random_prob = MIS.random_instance(
            n_nodes=10, edge_probability=0.8, seed=seed
        ).qubo

        self.assertTrue(
            terms_list_equality(mis_manual_prob.terms, mis_random_prob.terms)
        )
        self.assertEqual(mis_manual_prob.weights, mis_random_prob.weights)
        self.assertEqual(mis_manual_prob.constant, mis_random_prob.constant)

    def test_mis_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # graph type-check
        graph_list = [(1, 2), {"node1": 1, "node2": 2}, np.array([1, 2])]

        for each_graph in graph_list:
            with self.assertRaises(TypeError) as e:
                MIS(G=each_graph)
            self.assertEqual(
                "Input problem graph must be a networkx Graph.", str(e.exception)
            )

    def test_mis_classical_sol(self):
        """Test the maximal independent set random instance method classical solution"""

        seed = 1234
        np.random.seed(seed)
        mis_sol = MIS.random_instance(
            n_nodes=10, edge_probability=0.7, seed=seed
        ).classical_solution()

        sol = {
            "x_0": 1,
            "x_1": 1,
            "x_2": 0,
            "x_3": 0,
            "x_4": 0,
            "x_5": 0,
            "x_6": 0,
            "x_7": 0,
            "x_8": 1,
            "x_9": 0,
        }

        self.assertEqual(mis_sol, sol)

    def test_mis_plot(self):
        """Test maximal independent set random instance method"""
        from matplotlib.pyplot import Figure

        seed = 1234
        mis_random_prob = MIS.random_instance(
            n_nodes=10, edge_probability=0.7, seed=seed
        )
        sol = {
            "x_0": 1,
            "x_1": 1,
            "x_2": 0,
            "x_3": 0,
            "x_4": 0,
            "x_5": 0,
            "x_6": 0,
            "x_7": 0,
            "x_8": 1,
            "x_9": 0,
        }
        fig = mis_random_prob.plot_solution(sol)
        self.assertTrue(isinstance(fig, Figure))


if __name__ == "__main__":
    unittest.main()



from typing import List, Union, Optional, Tuple
import numpy as np

from .variational_baseparams import QAOAVariationalBaseParams
from .annealingparams import QAOAVariationalAnnealingParams
from .fourierparams import (
    QAOAVariationalFourierParams,
    QAOAVariationalFourierExtendedParams,
    QAOAVariationalFourierWithBiasParams,
)
from .extendedparams import QAOAVariationalExtendedParams
from .standardparams import (
    QAOAVariationalStandardParams,
    QAOAVariationalStandardWithBiasParams,
)
from ..ansatz_constructor import QAOADescriptor

VARIATIONAL_PARAMS_DICT_KEYS = {
    "standard": ["betas", "gammas"],
    "standard_w_bias": ["betas", "gammas_singles", "gammas_pairs"],
    "extended": ["betas_singles", "betas_pairs", "gammas_singles", "gammas_pairs"],
    "fourier": ["q", "v", "u"],
    "fourier_extended": ["q", "v_singles", "v_pairs", "u_singles", "u_pairs"],
    "fourier_w_bias": ["q", "v", "u_singles", "u_pairs"],
    "annealing": ["total_annealing_time", "schedule"],
}

PARAMS_CLASSES_MAPPER = {
    "standard": QAOAVariationalStandardParams,
    "standard_w_bias": QAOAVariationalStandardWithBiasParams,
    "extended": QAOAVariationalExtendedParams,
    "fourier": QAOAVariationalFourierParams,
    "fourier_extended": QAOAVariationalFourierExtendedParams,
    "fourier_w_bias": QAOAVariationalFourierWithBiasParams,
    "annealing": QAOAVariationalAnnealingParams,
}

SUPPORTED_INITIALIZATION_TYPES = ["rand", "ramp", "custom"]


def _qaoa_variational_params_args(
    params_type: str,
    init_type: str,
    betas: Optional[Union[List, np.ndarray]] = None,
    gammas: Optional[Union[List, np.ndarray]] = None,
    betas_singles: Optional[Union[List, np.ndarray]] = None,
    betas_pairs: Optional[Union[List, np.ndarray]] = None,
    gammas_singles: Optional[Union[List, np.ndarray]] = None,
    gammas_pairs: Optional[Union[List, np.ndarray]] = None,
    q: Optional[int] = None,
    v: Optional[Union[List, np.ndarray]] = None,
    u: Optional[Union[List, np.ndarray]] = None,
    v_singles: Optional[Union[List, np.ndarray]] = None,
    v_pairs: Optional[Union[List, np.ndarray]] = None,
    u_singles: Optional[Union[List, np.ndarray]] = None,
    u_pairs: Optional[Union[List, np.ndarray]] = None,
    total_annealing_time: Optional[float] = None,
    schedule: Optional[float] = None,
) -> Tuple[Union[List, np.ndarray]]:
    """
    Provided the given parameterisation type return the
    variational parameters arguments as a tuple

    Parameters
    ----------
    params_type: ``str``
        Chosen QAOA parameterisation strategy
    gammas: ``Union[List, np.ndarray]``
    betas: ``Union[List, np.ndarray]``
    gammas_singles: ``Union[List, np.ndarray]``
    gammas_pairs: ``Union[List, np.ndarray]``
    betas_singles: ``Union[List, np.ndarray]``
    betas_pairs: ``Union[List, np.ndarray]``
    q: ``int``
    v: ``Union[List, np.ndarray]``
    u: ``Union[List, np.ndarray]``
    v_singles: ``Union[List, np.ndarray]``
    v_pairs: ``Union[List, np.ndarray]``
    u_singles: ``Union[List, np.ndarray]``
    u_pairs: ``Union[List, np.ndarray]``
    total_annealing_time: ``float``
    schedule: ``Union[List, np.ndarray]``

    Returns
    -------
    variational_params_args: ``Tuple[Union[List,np.ndarray]]``
        Tuple object containing the correct angles for the specified parameterisation type
    """
    VARIATIONAL_PARAMS_MAPPER = {
        "standard": (betas, gammas),
        "standard_w_bias": (betas, gammas_singles, gammas_pairs),
        "extended": (betas_singles, betas_pairs, gammas_singles, gammas_pairs),
        "fourier": (q, v, u),
        "fourier_extended": (q, v_singles, v_pairs, u_singles, u_pairs),
        "fourier_w_bias": (q, v, u_singles, u_pairs),
        "annealing": (total_annealing_time, schedule),
    }
    final_variational_params = VARIATIONAL_PARAMS_MAPPER[params_type.lower()]
    if init_type != "custom":
        final_variational_params = tuple(
            param for param in final_variational_params if param is not None
        )

    return final_variational_params


def create_qaoa_variational_params(
    qaoa_descriptor: QAOADescriptor,
    params_type: str,
    init_type: str,
    variational_params_dict: dict = {},
    linear_ramp_time: Optional[float] = None,
    q: Optional[int] = None,
    total_annealing_time: Optional[float] = None,
    seed: int = None,
) -> QAOAVariationalBaseParams:
    """
    Create QAOA Variational Parameters of the specified type.

    Parameters
    ----------
    qaoa_descriptor: ``QAOADescriptor``
        QAOA Circuit Parameters containing information on
        mixer and cost Hamiltonians.
    params_type: ``str``
        Type of variational parameters to be created.
    init_type: ``str``
        Type of initialisation to be used.
    variational_params_dict: ``dict``
        Dictionary containing the variational angles. It can only have the following
        keys (with corresponding angle values):
        For example, {'betas': [0.1, 0.2, 0.3], 'gammas': [0.1, 0.2, 0.3]}
    linear_ramp_time: ``float``
        Total time for the linear ramp initialisation method for parameters.
    Returns
    -------
        QAOAVariationalBaseParams object of the required type
    """

    params_type = params_type.lower()
    init_type = init_type.lower()
    variational_params_dict.update({"q": q})
    variational_params_dict.update({"total_annealing_time": total_annealing_time})
    variational_params_args = _qaoa_variational_params_args(
        params_type, init_type, **variational_params_dict
    )

    try:
        params_class = PARAMS_CLASSES_MAPPER[params_type]
    except KeyError:
        raise ValueError(
            f"{params_type} Parameterisation is not supported."
            f" Choose from {PARAMS_CLASSES_MAPPER.keys()}"
        )

    if init_type == "custom":
        try:
            qaoa_variational_params = params_class(
                qaoa_descriptor, *variational_params_args
            )
        except:
            raise ValueError(
                f"For the selected {params_type} parameterisation, please specify a"
                f" dictionary with correct {VARIATIONAL_PARAMS_DICT_KEYS[params_type]} keys"
            )
    elif init_type == "ramp":
        if isinstance(linear_ramp_time, float) or isinstance(linear_ramp_time, int):
            assert (
                linear_ramp_time > 0
            ), "Please specify the linear ramp time. Only positive values are allowed."
        elif linear_ramp_time is None:
            pass
        else:
            raise ValueError(f"Please specify a numeric value for linear_ramp_time.")

        qaoa_variational_params = params_class.linear_ramp_from_hamiltonian(
            qaoa_descriptor, *variational_params_args, time=linear_ramp_time
        )
    elif init_type == "rand":
        qaoa_variational_params = params_class.random(
            qaoa_descriptor, *variational_params_args, seed=seed
        )
    else:
        raise ValueError(
            f"{init_type} Initialisation strategy is not yet supported."
            f" Please choose from {SUPPORTED_INITIALIZATION_TYPES}"
        )

    return qaoa_variational_params


def qaoa_variational_params_converter(
    target_params_type: str, current_params_obj: QAOAVariationalBaseParams
) -> QAOAVariationalBaseParams:
    """
    Convert the current variational parameters object to the target variational parameters object.

    Parameters
    ----------
    target_params_type: ``str``
        Type of variational parameters to be created.
    current_params_obj: ``QAOAVariationalBaseParams``
        Current variational parameters object.

    Returns
    -------
    converted_params_obj: ``QAOAVariationalBaseParams``
        Converted variational parameters object.
    """

    try:
        new_params_class = PARAMS_CLASSES_MAPPER[target_params_type]
    except KeyError:
        raise ValueError(
            f"{target_params_type} Parameterisation is not supported."
            f" Choose from {PARAMS_CLASSES_MAPPER.keys()}"
        )

    converted_params_obj = new_params_class.from_other_parameters(current_params_obj)

    return converted_params_obj



from collections import defaultdict
from typing import Union

import numpy as np

from .qubo import QUBO


class FromDocplex2IsingModel(object):
    def __init__(
        self,
        model,
        multipliers: Union[float, list] = None,
        unbalanced_const: bool = False,
        strength_ineq: list = [0.1, 0.5],
    ):
        """
        Creates an instance to translate Docplex models to its Ising Model representation

        Parameters
        ----------
        model : Docplex model
            It is an object that has the mathematical expressions (cost function and
            inequality and Equality constraints) of an optimization problem in
            a binary representation.
        multipliters: [float, integer, list]
            The strength of the penalties of the cost function
        heuristic: bool
            If the method for the inequality constraints is used. This method
            implement a novel approach that do not required slack variables.
        strength_ineq: List[float, float]
            Lagrange multipliers of the penalization term using the unbalanced
            penalization method.
            For the unbalanced penalization => - \lambda_1 h(x) + \lambda_2 * h(x)**2
            where h(x) >= 0 is the inequality constraint.
            strength_ineq = [\lambda_1, \lambda_2]
            Usually \lambda_2 < \lambda_1. Please refer to the paper:
                Unbalanced penalization: A new approach to encode inequality constraints
                of combinatorial problems for quantum optimization algorithms
                https://arxiv.org/abs/2211.13914

        """
        # assign the docplex Model
        self.model = model.copy()
        self.model_name = model.name
        # save the index in a dict
        self.idx_terms = {}
        for x in self.model.iter_variables():
            self.idx_terms[x] = x.index
            if x.vartype.short_name != "binary":
                TypeError(f"Variable {x.vartype.short_name} is not allowed.")
        # if unbalanced constraint  approach
        self.unbalanced = unbalanced_const
        self.strength_ineq = strength_ineq
        # get doclex qubo and ising model
        self.qubo_docplex, self.ising_model = self.get_models(multipliers)

    def linear_expr(self, expr):
        """
        Adds linear expresions to the objective function stored in self.qubo_dict.

        Parameters
        ----------
        expr : model.objective_expr
            A docplex model attribute.

        """
        for x, weight in expr.get_linear_part().iter_terms():
            self.qubo_dict[(self.idx_terms[x],)] += weight

    def quadratic_expr(self, expr):
        """
        Adds quadratic terms to the objective function stored in self.qubo_dict.

        Parameters
        ----------
        expr : model.objective_expr
            A docplex model attribute.
        """

        # save the terms and coeffs of the quadratic part,
        # considering two index i and j
        for x, y, weight in expr.iter_quad_triplets():
            i = self.idx_terms[x]
            j = self.idx_terms[y]
            if i == j:  # if this is the term is  Z1**2 for example
                self.qubo_dict[(i,)] += weight
            else:
                self.qubo_dict[(i, j)] += weight

    @staticmethod
    def equality_to_penalty(expression, multiplier: Union[int, float]):
        """
        Add equality constraints to the cost function using the penality representation.
        The constraints should be linear.

        Parameters
        ----------
        expression : docplex.mp.linear.LinearExpr
            docplex equality constraint "expression == 0".
        multiplier : float
            Lagrange multiplier of the penalty

        Returns
        -------
        penalty : docplex.mp.quad.QuadExpr
            Penalty that will be added to the cost function.

        """
        penalty = multiplier * (expression) ** 2
        return penalty

    @staticmethod
    def bounds(linear_expression):
        """
        Generates the limits of a linear term

        Parameters
        ----------
        linear_exp : generator
            Iterable term with the quadratic terms of the cost function.

        Returns
        -------
        l_bound : float
            Lower bound of the quadratic term.
        u_bound : float
            Upper bound of the quadratic term
        """

        l_bound = (
            u_bound
        ) = linear_expression.constant  # Lower and upper bound of the constraint
        for term, coeff in linear_expression.iter_terms():
            l_bound += min(0, coeff)
            u_bound += max(0, coeff)

        return l_bound, u_bound

    @staticmethod
    def quadratic_bounds(iter_exp):
        """
        Generates the limits of the quadratic terms

        Parameters
        ----------
        iter_exp : generator
            Iterable term with the quadratic terms of the cost function.

        Returns
        -------
        l_bound : float
            Lower bound of the quadratic term.
        u_bound : float
            Upper bound of the quadratic term

        """
        l_bound = 0
        u_bound = 0
        for z1, z2, coeff in iter_exp:
            l_bound += min(0, coeff)
            u_bound += max(0, coeff)
        return l_bound, u_bound

    def inequality_to_equality(self, constraint):
        """
        Transform inequality contraints into equality constriants using
        slack variables.

        Parameters
        ----------
        constraint : docplex.mp.constr.LinearConstraint
            DESCRIPTION.

        Returns
        -------
        new_exp : docplex.mp.linear.LinearExpr
            The equality constraint representation of the inequality constraint.

        """

        if constraint.sense_string == "LE":  # Less or equal inequality constraint
            new_exp = constraint.get_right_expr() + -1 * constraint.get_left_expr()
        elif constraint.sense_string == "GE":  # Great or equal inequality constriant
            new_exp = constraint.get_left_expr() + -1 * constraint.get_right_expr()
        else:
            raise AttributeError(
                f"It is not possible to implement constraint {constraint.sense_string}."
            )

        lower_bound, upper_bound = self.bounds(new_exp)
        slack_lim = upper_bound  # Slack var limit

        if slack_lim > 0:
            sign = -1
        else:
            sign = 1
        # Create the slack variables to represent valid terms in the inequality
        n_slack = int(np.ceil(np.log2(abs(slack_lim + 1))))
        if n_slack > 0:
            slack_vars = self.model.binary_var_list(
                n_slack, name=f"slack_{constraint.name}"
            )
            for x in slack_vars:
                self.idx_terms[x] = x.index

            for nn, var in enumerate(slack_vars[:-1]):
                new_exp += sign * (2**nn) * var

            new_exp += (
                sign * (slack_lim - 2 ** (n_slack - 1) + 1) * slack_vars[-1]
            )  # restrict the last term to fit the upper bound

        return new_exp

    def inequality_to_unbalanced_penalty(self, constraint):
        """
        Inequality constraint based on an unbalanced penality function described in
        detail in the paper: https://arxiv.org/abs/2211.13914

        Parameters
        ----------
        constraint : DOcplex inequality constraint
            Inequality constraints in a DOcplex format.

        Returns
        -------
        penalty : DOcplex term
            Quadratic programing penalization term.

        """
        if constraint.sense_string == "LE":  # Less or equal inequality constraint
            new_exp = constraint.get_right_expr() + -1 * constraint.get_left_expr()
        elif constraint.sense_string == "GE":  # Great or equal inequality constriant
            new_exp = constraint.get_left_expr() + -1 * constraint.get_right_expr()
        else:
            raise AttributeError(
                f"It is not possible to implement constraint {constraint.sense_string}."
            )
        strength = self.strength_ineq
        penalty = -strength[0] * new_exp + strength[1] * new_exp**2
        return penalty

    def multipliers_generators(self):
        """
        Penality term size adapter, this is the Lagrange multiplier of the cost
        function penalties for every constraint if the multiplier is not indicated
        by the user.

        Returns
        -------
        float
            the multiplier resized by the cost function limits.

        """
        cost_func = self.model.objective_expr
        l_bound_linear, u_bound_linear = self.bounds(cost_func.get_linear_part())
        l_bound_quad, u_bound_quad = self.quadratic_bounds(
            cost_func.iter_quad_triplets()
        )
        return 1.0 + (u_bound_linear - l_bound_linear) + (u_bound_quad - l_bound_quad)

    def linear_constraints(self, multipliers=None) -> None:
        """
        Adds the constraints of the problem to the objective function.

        Parameters
        ----------
        multiplier : List
            For each constraint a multiplier, if None it automatically is selected.

        Returns
        -------
        None.

        """

        constraints_list = list(self.model.iter_linear_constraints())
        n_constraints = len(constraints_list)

        if (
            multipliers is None
        ):  # Default penalties are choosen from the bounds of the objective func.
            multipliers = n_constraints * [self.multipliers_generators()]

        elif np.isscalar(multipliers):
            multipliers = n_constraints * [multipliers]

        elif type(multipliers) not in [list, np.ndarray]:
            raise TypeError(f"{type(multipliers)} is not a accepted format")

        for cn, constraint in enumerate(constraints_list):
            if (
                constraint.sense_string == "EQ"
            ):  # Equality constraint added as a penalty.
                left_exp = constraint.get_left_expr()
                right_exp = constraint.get_right_expr()
                expression = left_exp + -1 * right_exp
                penalty = self.equality_to_penalty(expression, multipliers[cn])
            elif constraint.sense_string in [
                "LE",
                "GE",
            ]:  # Inequality constraint added as a penalty with additional slack variables.
                constraint.name = f"C{cn}"
                if self.unbalanced:
                    penalty = self.inequality_to_unbalanced_penalty(constraint)
                else:
                    ineq2eq = self.inequality_to_equality(constraint)
                    penalty = self.equality_to_penalty(ineq2eq, multipliers[cn])
            else:
                raise TypeError("This is not a valid constraint.")

            self.linear_expr(penalty)
            self.quadratic_expr(penalty)
            self.constant += penalty.constant
            self.objective_qubo += penalty

    @staticmethod
    def qubo_to_ising(n_variables, qubo_terms, qubo_weights):
        """
        Converts the terms and weights in QUBO representation ([0,1])
        to the Ising representation ([-1, 1]).

        Parameters
        ----------
        n_variables : int
            number of variables.
        qubo_terms : List
            List of QUBO variables.
        qubo_weights : List
            coefficients of the variables


        Returns
        -------
        Ising Model stored on QUBO class

        """
        ising_terms, ising_weights = [], []
        linear_terms = np.zeros(n_variables)

        constant_term = 0
        # Process the given terms and weights
        for weight, term in zip(qubo_weights, qubo_terms):
            if len(term) == 2:
                u, v = term

                if u != v:
                    ising_terms.append([u, v])
                    ising_weights.append(weight / 4)
                else:
                    constant_term += weight / 4

                linear_terms[term[0]] -= weight / 4
                linear_terms[term[1]] -= weight / 4
                constant_term += weight / 4
            elif len(term) == 1:
                linear_terms[term[0]] -= weight / 2
                constant_term += weight / 2
            elif len(term) == 0:
                constant_term += weight
            else:
                raise TypeError(f"Term {term} is not recognized!")

        for variable, linear_term in enumerate(linear_terms):
            ising_terms.append([variable])
            ising_weights.append(linear_term)

        ising_terms.append([])
        ising_weights.append(constant_term)

        return QUBO(n_variables, ising_terms, ising_weights)

    def get_models(self, multipliers: Union[float, list] = None):
        """
        Creates a QUBO docplex model, QUBO dict OQ model, and an Ising Model form
        a Docplex quadratic program.

        Parameters
        ----------
        model : Docplex model
            It is an object that has the mathematical expressions (Cost function,
            Inequality and Equality constraints) of an optimization problem.

        Returns
        -------
        qubo_docplex, ising_model


        """
        # save a dictionary with the qubo information
        self.qubo_dict = defaultdict(float)

        # objective sense
        if self.model.objective_sense.is_minimize():
            # If it is minimized
            self.objective_expr = self.model.objective_expr
        else:
            # If it is maximized, multiplied by -1 to minimize.
            self.objective_expr = -1 * self.model.objective_expr

        # Objective QUBO
        self.objective_qubo = self.objective_expr
        # Obtain the constant from the model
        self.constant = self.objective_expr.constant

        # Save the terms and coeffs from the linear part
        self.linear_expr(self.objective_expr)

        # Save the terms and coeffs from the quadratic part
        self.quadratic_expr(self.objective_expr)

        # Add the linear constraints into the qubo
        self.linear_constraints(multipliers=multipliers)

        terms = list(self.qubo_dict.keys()) + [
            []
        ]  # The right term is for adding the constant part of the QUBO

        weights = list(self.qubo_dict.values()) + [self.constant]

        n_variables = self.model.number_of_variables
        # QUBO docplex
        qubo_docplex = self.model.copy()
        qubo_docplex.name = self.model_name
        qubo_docplex.clear_constraints()
        qubo_docplex.remove_objective()
        qubo_docplex.minimize(self.objective_qubo)
        # Ising Hamiltonian of the QUBO
        ising_model = self.qubo_to_ising(n_variables, terms, weights)

        return qubo_docplex, ising_model



import warnings
import numpy as np
import unittest

# OpenQAOA imports
from openqaoa.backends import QAOAvectorizedBackendSimulator
from openqaoa.qaoa_components import (
    QAOADescriptor,
    Hamiltonian,
    create_qaoa_variational_params,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.optimizers.logger_vqa import Logger
from openqaoa.derivatives.derivative_functions import derivative

"""
Unittest based testing of derivative computations.
"""


class TestQAOACostBaseClass(unittest.TestCase):

    """
    def test_gradient_agreement(self):
        "Test agreement between gradients computed from finite difference, parameter shift and SPS (all gates sampled) for weighted and unweighted graphs at several parameters."

        # unweighted graph
        terms = [[0,1], [0,2], [1,3], [2]]
        weights = [1, 1, 1, 1]
        register = [0, 1, 2, 3]
        p = 2
        nqubits = 4

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(
            terms, weights, constant=0)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = create_qaoa_variational_params(
            qaoa_qaoa_descriptor, 'standard', 'ramp')
        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True)

        grad_stepsize = 0.00000001
        gradient_ps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'param_shift')
        gradient_fd = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'finite_difference', {'stepsize': grad_stepsize})
        gradient_sps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'stoch_param_shift', {'stepsize':grad_stepsize, 'n_beta_single':-1, 'n_beta_pair':-1, 'n_gamma_pair':-1, 'n_gamma_single':-1})

        params = [[0,0,0,0], [1,1,1,1], [np.pi/2, np.pi/2, np.pi/2, np.pi/2]]

        for param in params:
            grad_fd = gradient_fd(param)
            grad_ps = gradient_ps(param)
            grad_sps = gradient_sps(param)

            for i, grad in enumerate(grad_fd):
                assert np.isclose(grad, grad_ps[i], rtol=1e-05, atol=1e-05)
                assert np.isclose(grad, grad_sps[i], rtol=1e-05, atol=1e-05)

        # weighted graph with bias
        terms = [[0,1], [1,2], [0,3], [2], [1]]
        weights = [1, 1.1, 1.5, 2, -0.8]
        register = [0, 1, 2, 3]
        p = 2
        nqubits = 4

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(
            terms, weights, constant=0.8)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = create_qaoa_variational_params(
            qaoa_qaoa_descriptor, 'standard', 'ramp')
        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True)

        grad_stepsize = 0.00000001
        gradient_ps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'param_shift')
        gradient_fd = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'finite_difference', {'stepsize': grad_stepsize})
        gradient_sps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'stoch_param_shift', {'stepsize':grad_stepsize, 'n_beta_single':-1, 'n_beta_pair':-1, 'n_gamma_pair':-1, 'n_gamma_single':-1})

        params = [[0,0,0,0], [1,1,1,1], [np.pi/2, np.pi/2, np.pi/2, np.pi/2]]

        for param in params:
            grad_fd = gradient_fd(param)
            grad_ps = gradient_ps(param)
            grad_sps = gradient_sps(param)

            for i, grad in enumerate(grad_fd):
                assert np.isclose(grad, grad_ps[i], rtol=1e-05, atol=1e-05)
                assert np.isclose(grad, grad_sps[i], rtol=1e-05, atol=1e-05)
    """

    def setUp(self):
        self.log = Logger(
            {
                "func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "jac_func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "n_shots": {
                    "history_update_bool": True,
                    "best_update_string": "Replace",
                },
            },
            {
                "root_nodes": ["func_evals", "jac_func_evals", "n_shots"],
                "best_update_structure": [],
            },
        )

        self.log.log_variables({"func_evals": 0})
        self.log.log_variables({"jac_func_evals": 0})

    def __backend_params(self, terms: list, weights: list, p: int, nqubits: int):
        cost_hamiltonian = Hamiltonian.classical_hamiltonian(terms, weights, constant=0)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = create_qaoa_variational_params(
            qaoa_qaoa_descriptor, "standard", "ramp"
        )
        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_qaoa_descriptor,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
        )
        return backend_vectorized, variational_params_std

    def test_gradient_computation(self):
        "Test gradient computation by param. shift, finite difference, and SPS (all gates sampled) on barbell graph."

        # Analytical cost expression : C(b,g) = -sin(4b)*sin(2g)
        backend, params = self.__backend_params(
            terms=[[0, 1]], weights=[1], p=1, nqubits=2
        )

        gradients_types_list = ["finite_difference", "param_shift", "stoch_param_shift"]
        gradients_fun_list = [
            derivative(
                backend, params, self.log, "gradient", type_, {"stepsize": 0.0000001}
            )
            for type_ in gradients_types_list
        ]

        test_points = [[0, 0], [np.pi / 2, np.pi / 3], [1, 2]]

        for point in test_points:
            beta, gamma = point

            dCdb, dCdg = -4 * np.cos(4 * beta) * np.sin(2 * gamma), -2 * np.sin(
                4 * beta
            ) * np.cos(2 * gamma)

            for gradient_fun, gradient_name in zip(
                gradients_fun_list, gradients_types_list
            ):
                dCdb_, dCdg_ = gradient_fun(point)
                assert np.isclose(
                    dCdb, dCdb_, rtol=1e-05, atol=1e-05
                ), f"Gradient computation failed for {gradient_name} on barbell graph. dCdb: {dCdb}, dCdb_: {dCdb_}"
                assert np.isclose(
                    dCdg, dCdg_, rtol=1e-05, atol=1e-05
                ), f"Gradient computation failed for {gradient_name} on barbell graph. dCdg: {dCdg}, dCdg_: {dCdg_}"

    def test_gradient_w_variance_computation(self):
        "Test gradient computation by param. shift, finite difference, and SPS (all gates sampled) on barbell graph."

        # Analytical cost expression : C(b,g) = -sin(4b)*sin(2g)
        backend, params = self.__backend_params(
            terms=[[0, 1]], weights=[1], p=1, nqubits=2
        )

        gradients_types_list = ["finite_difference", "param_shift", "stoch_param_shift"]
        gradients_fun_list = [
            derivative(
                backend,
                params,
                self.log,
                "gradient_w_variance",
                type_,
                {"stepsize": 0.0000001},
            )
            for type_ in gradients_types_list
        ]

        test_points = [[0, 0], [np.pi / 2, np.pi / 3], [1, 2]]

        for point in test_points:
            for gradient_fun, gradient_name in zip(
                gradients_fun_list, gradients_types_list
            ):
                # compute gradient for each point with number of shots 1000 for ech function evaluation
                grad, var, n_shots = gradient_fun(point, n_shots=1000)

                # check if there is a gradient and variance (we can't check the value of the gradient and variance because it is randomly computed, since n_shots is 1000)
                for g in grad:
                    assert (
                        np.abs(g) >= 0
                    ), f"Gradient computation failed for {gradient_name} on barbell graph. grad: {grad}"
                for v in var:
                    assert (
                        v > 0
                    ), f"Error computing the variance of the gradient for {gradient_name} on barbell graph."

                # check if the number of shots is correct
                x = (
                    4 if gradient_name == "finite_difference" else 6
                )  # that is also checking that  SPS samples all gates when (n_beta, n_gamma_pair, n_gamma_single) is (-1, -1, -1)
                assert (
                    n_shots == x * 1000
                ), f"The number of shots should be {x*1000} but is {n_shots}."

    def test_SPS_sampling(self):
        "Test that SPS samples the number of gates specified by the user."
        backend, params = self.__backend_params(
            terms=[[0, 1]], weights=[1], p=1, nqubits=2
        )
        gradient_fun = derivative(
            backend,
            params,
            self.log,
            "gradient_w_variance",
            "stoch_param_shift",
            {
                "stepsize": 0.1,
                "n_beta_single": 1,
                "n_beta_pair": 0,
                "n_gamma_pair": 1,
                "n_gamma_single": 0,
            },
        )

        n_shots = gradient_fun([0, 0], n_shots=1000)[2]
        # check if the number of shots is correct, we know that we are sampling 2 gates (1 beta and 1 gamma) -> 2*2*1000 (the second 2* comes from two evaluations for each gradient)
        assert (
            n_shots == 4000
        ), f"The number of shots should be {4000} but is {n_shots}."

    def test_hessian_computation(self):
        "Test Hessian computation by finite difference on barbell graph"

        # Analytical cost expression : C(b,g) = -sin(4b)*sin(2g)
        backend, params = self.__backend_params(
            terms=[[0, 1]], weights=[1], p=1, nqubits=2
        )

        hessian_fd = derivative(
            backend,
            params,
            self.log,
            "hessian",
            "finite_difference",
            {"stepsize": 0.001},
        )

        test_points = np.round([[0, 0], [np.pi / 2, np.pi / 3], [1, 2]], 12)

        for point in test_points:
            beta, gamma = point[0], point[1]

            dCdbb = 16 * np.sin(4 * beta) * np.sin(2 * gamma)
            dCdbg = -8 * np.cos(4 * beta) * np.cos(2 * gamma)
            dCdgb = dCdbg
            dCdgg = 4 * np.sin(4 * beta) * np.sin(2 * gamma)

            assert np.isclose(dCdbb, hessian_fd(point)[0][0], rtol=1e-05, atol=1e-05)
            assert np.isclose(dCdbg, hessian_fd(point)[0][1], rtol=1e-05, atol=1e-05)
            assert np.isclose(dCdgb, hessian_fd(point)[1][0], rtol=1e-05, atol=1e-05)
            assert np.isclose(dCdgg, hessian_fd(point)[1][1], rtol=1e-05, atol=1e-05)


if __name__ == "__main__":
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=PendingDeprecationWarning)
        unittest.main()



import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import MaximumCut


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool


class TestMaximumCut(unittest.TestCase):
    """Tests for MaximumCut class"""

    def test_maximumcut_terms_weights_constant(self):
        """Test that MaximumCut creates a correct QUBO from the provided graph"""

        gr = nx.generators.random_graphs.fast_gnp_random_graph(n=10, p=0.8)
        gr_edges = [list(edge) for edge in gr.edges()]
        gr_weights = [1] * len(gr_edges)

        maxcut_prob_qubo = MaximumCut(gr).qubo

        self.assertTrue(terms_list_equality(gr_edges, maxcut_prob_qubo.terms))
        self.assertEqual(gr_weights, maxcut_prob_qubo.weights)
        self.assertEqual(0, maxcut_prob_qubo.constant)

    def test_maximumcut_random_problem(self):
        """Test MaximumCut random instance method"""

        seed = 1234
        gr = nx.generators.random_graphs.fast_gnp_random_graph(n=10, p=0.8, seed=seed)
        maxcut_manual_prob = MaximumCut(gr).qubo

        np.random.seed(1234)
        maxcut_random_prob = MaximumCut.random_instance(
            n_nodes=10, edge_probability=0.8, seed=seed
        ).qubo

        self.assertTrue(
            terms_list_equality(maxcut_manual_prob.terms, maxcut_random_prob.terms)
        )
        self.assertEqual(maxcut_manual_prob.weights, maxcut_random_prob.weights)
        self.assertEqual(maxcut_manual_prob.constant, maxcut_random_prob.constant)

    def test_maximumcut_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # graph type-check
        graph_list = [(1, 2), {"node1": 1, "node2": 2}, np.array([1, 2])]

        for each_graph in graph_list:
            with self.assertRaises(TypeError) as e:
                MaximumCut(G=each_graph)
            self.assertEqual(
                "Input problem graph must be a networkx Graph.", str(e.exception)
            )


if __name__ == "__main__":
    unittest.main()



from typing import Union

from ..workflow_properties import WorkflowProperties
from ...qaoa_components import Hamiltonian


ALLOWED_RQAOA_TYPES = ["adaptive", "custom"]


class RqaoaParameters(WorkflowProperties):
    """
    Parameter class to initialise parameters to run a desired RQAOA program.

    Attributes
    ----------
    rqaoa_type: `int`
        String specifying the RQAOA scheme under which eliminations are computed. The two methods are 'custom' and
        'adaptive'. Defaults to 'adaptive'.
    n_max: `int`
        Maximum number of eliminations allowed at each step when using the adaptive method.
    steps: `Union[list,int]`
        Elimination schedule for the RQAOA algorithm. If an integer is passed, it sets the number of spins eliminated
        at each step. If a list is passed, the algorithm will follow the list to select how many spins to eliminate
        at each step. Note that the list needs enough elements to specify eliminations from the initial number of qubits
        up to the cutoff value. If the list contains more, the algorithm will follow instructions until the cutoff value
        is reached.
    n_cutoff: `int`
        Cutoff value at which the RQAOA algorithm obtains the solution classically.
    original_hamiltonian: `Hamiltonian`
        Hamiltonian encoding the original problem fed into the RQAOA algorithm.
    counter: `int`
        Variable to count the step in the schedule. If counter = 3 the next step is schedule[3].
        Default is 0, but can be changed to start in the position of the schedule that one wants.
    """

    def __init__(
        self,
        rqaoa_type: str = "custom",
        n_max: int = 1,
        steps: Union[list, int] = 1,
        n_cutoff: int = 5,
        original_hamiltonian: Hamiltonian = None,
        counter: int = 0,
    ):
        """
        Initialises RQAOA program parameters.

        Parameters
        ----------
        rqaoa_type: `int`
            String specifying the RQAOA scheme under which eliminations are computed. The two methods are 'custom' and
            'adaptive'. Defaults to "custom".
        n_max: `int`
            Maximum number of eliminations allowed at each step when using the adaptive method. Defaults to 1.
        steps: `Union[list,int]`
            Elimination schedule for the RQAOA algorithm. If an integer is passed, it sets the number of spins
            eliminated at each step. If a list is passed, the algorithm will follow the list to select
            how many spins to eliminate at each step. Note that the list needs enough elements to specify
            eliminations from the initial number of qubits up to the cutoff value. If the list contains more,
            the algorithm will follow instructions until the cutoff value
            is reached. Defaults to 1.
        n_cutoff: `int`
            Cutoff value at which the RQAOA algorithm obtains the solution classically. Defaults to 5.
        original_hamiltonian: `Hamiltonian`
            Hamiltonian encoding the original problem fed into the RQAOA algorithm. Defaults to None.
        counter: `int`
            Variable to count the step in the schedule. If counter = 3 the next step is schedule[3].
            Default is 0, but can be changed to start in the position of the schedule that one wants.
        """
        self.rqaoa_type = rqaoa_type.lower()
        self.n_max = n_max
        self.steps = steps
        self.n_cutoff = n_cutoff
        self.original_hamiltonian = original_hamiltonian
        self.counter = counter

        # check if the rqaoa type is correct
        if self.rqaoa_type not in ALLOWED_RQAOA_TYPES:
            self.compiled = False
            raise Exception(
                f'rqaoa_type {self.rqaoa_type} is not supported. Please select "adaptive" or "custom".'
            )

        # check if the parameters given are in accordance with the rqaoa_type
        if self.rqaoa_type == "adaptive":
            if self.steps != 1:
                raise ValueError(
                    "When using the adaptive method, the `steps` parameter is not required.  \
                    The parameter that specifies the maximum number of eliminations per step is `n_max`."
                )
            if self.counter != 0:
                raise ValueError(
                    "When using the adaptive method, the `counter` parameter is not required."
                )
        else:
            if self.n_max != 1:
                raise ValueError(
                    "When using the custom method, the `n_max` parameter is not required.  \
                    The parameter that specifies the number of eliminations is `steps`,\
                          which can be a string or a list."
                )



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""Adam optimizer"""
from numpy import sqrt  # changed from the original PennyLane code
from .gradient_descent import GradientDescentOptimizer


class AdamOptimizer(GradientDescentOptimizer):
    r"""Gradient-descent optimizer with adaptive learning rate, first and second moment.

    Adaptive Moment Estimation uses a step-dependent learning rate,
    a first moment :math:`a` and a second moment :math:`b`, reminiscent of
    the momentum and velocity of a particle:

    .. math::
        x^{(t+1)} = x^{(t)} - \eta^{(t+1)} \frac{a^{(t+1)}}{\sqrt{b^{(t+1)}} + \epsilon },

    where the update rules for the two moments are given by

    .. math::
        a^{(t+1)} &= \beta_1 a^{(t)} + (1-\beta_1) \nabla f(x^{(t)}),\\
        b^{(t+1)} &= \beta_2 b^{(t)} + (1-\beta_2) (\nabla f(x^{(t)}))^{\odot 2},\\
        \eta^{(t+1)} &= \eta \frac{\sqrt{(1-\beta_2^{t+1})}}{(1-\beta_1^{t+1})}.

    Above, :math:`( \nabla f(x^{(t-1)}))^{\odot 2}` denotes the element-wise square operation,
    which means that each element in the gradient is multiplied by itself. The hyperparameters
    :math:`\beta_1` and :math:`\beta_2` can also be step-dependent. Initially, the first and
    second moment are zero.

    The shift :math:`\epsilon` avoids division by zero.

    For more details, see `arXiv:1412.6980 <https://arxiv.org/abs/1412.6980>`_.

    Args:
        stepsize (float): the user-defined hyperparameter :math:`\eta`
        beta1 (float): hyperparameter governing the update of the first and second moment
        beta2 (float): hyperparameter governing the update of the first and second moment
        eps (float): offset :math:`\epsilon` added for numerical stability

    """

    def __init__(self, stepsize=0.01, beta1=0.9, beta2=0.99, eps=1e-8):
        super().__init__(stepsize)
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self.accumulation = None

    def apply_grad(self, grad, args):
        r"""Update the variables args to take a single optimization step. Flattens and unflattens
        the inputs to maintain nested iterables as the parameters of the optimization.

        Args:
            grad (tuple[ndarray]): the gradient of the objective
                function at point :math:`x^{(t)}`: :math:`\nabla f(x^{(t)})`
            args (tuple): the current value of the variables :math:`x^{(t)}`

        Returns:
            list: the new values :math:`x^{(t+1)}`
        """
        args_new = list(args)

        if self.accumulation is None:
            self.accumulation = {"fm": [0] * len(args), "sm": [0] * len(args), "t": 0}

        self.accumulation["t"] += 1

        # Update step size (instead of correcting for bias)
        new_stepsize = (
            self.stepsize
            * sqrt(1 - self.beta2 ** self.accumulation["t"])
            / (1 - self.beta1 ** self.accumulation["t"])
        )

        trained_index = 0
        for index, arg in enumerate(args):
            if getattr(arg, "requires_grad", False):
                self._update_accumulation(index, grad[trained_index])
                args_new[index] = arg - new_stepsize * self.accumulation["fm"][
                    index
                ] / (sqrt(self.accumulation["sm"][index]) + self.eps)

                trained_index += 1

        return args_new

    def _update_accumulation(self, index, grad):
        r"""Update the moments.

        Args:
            index (int): the index of the argument to update
            grad (ndarray): the gradient for that trainable param
        """
        # update first moment
        self.accumulation["fm"][index] = (
            self.beta1 * self.accumulation["fm"][index] + (1 - self.beta1) * grad
        )

        # update second moment
        self.accumulation["sm"][index] = (
            self.beta2 * self.accumulation["sm"][index] + (1 - self.beta2) * grad**2
        )

    def reset(self):
        """Reset optimizer by erasing memory of past steps."""
        self.accumulation = None

    @property
    def fm(self):
        """Returns estimated first moments of gradient"""
        return None if self.accumulation is None else self.accumulation["fm"]

    @property
    def sm(self):
        """Returns estimated second moments of gradient"""
        return None if self.accumulation is None else self.accumulation["sm"]

    @property
    def t(self):
        """Returns accumulated timesteps"""
        return None if self.accumulation is None else self.accumulation["t"]



import abc
import logging

logging.getLogger().setLevel(logging.ERROR)


class DeviceBase(metaclass=abc.ABCMeta):
    """An object that contains the relevant information required to access
    certain backends. Other Access Objects have to inherit from this object.
    """

    @abc.abstractmethod
    def check_connection(self) -> bool:
        """This method should allow a user to easily check if the credentials
        provided to access the remote QPU is valid.

        Returns
        -------
        bool
            True if a connection can be established. If False, the error
            should be logged and printable. (Not creating an exception
            here is good for extendibility. i.e. able to try multiple
            providers without exiting the program.)
        """
        pass

    @abc.abstractmethod
    def connectivity(self):
        """
        obtain the device connectivity as a list of qubit pairs

        Returns
        -------
            List[List[int]]
        """
        pass



import unittest
import numpy as np
from openqaoa.problems import Knapsack, SlackFreeKnapsack


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool


class TestKnapsack(unittest.TestCase):
    """Tests for Knapsack class"""

    def test_knapsack_terms_weights_constant(self):
        """Test that Knapsack creates the correct QUBO problem"""

        values = [2, 4, 3, 5]
        weights = [3, 6, 9, 1]
        weight_capacity = 15
        n_qubits = len(values) + int(np.ceil(np.log2(weight_capacity)))
        penalty = 2 * max(values)
        knap_terms = [
            [0, 1],
            [0, 2],
            [0, 3],
            [1, 2],
            [1, 3],
            [2, 3],
            [4, 5],
            [4, 6],
            [4, 7],
            [5, 6],
            [5, 7],
            [6, 7],
            [0, 4],
            [0, 5],
            [0, 6],
            [0, 7],
            [1, 4],
            [1, 5],
            [1, 6],
            [1, 7],
            [2, 4],
            [2, 5],
            [2, 6],
            [2, 7],
            [3, 4],
            [3, 5],
            [3, 6],
            [3, 7],
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
        ]
        knap_weights = [
            10.0,
            20.0,
            40.0,
            40.0,
            80.0,
            160.0,
            90.0,
            135.0,
            15.0,
            270.0,
            30.0,
            45.0,
            15.0,
            30.0,
            45.0,
            5.0,
            30.0,
            60.0,
            90.0,
            10.0,
            60.0,
            120.0,
            180.0,
            20.0,
            120.0,
            240.0,
            360.0,
            40.0,
            -20.0,
            -40.0,
            -80.0,
            -160.0,
            -59.0,
            -118.0,
            -178.5,
            -17.5,
        ]
        knap_constant = 563.0

        knapsack_prob_qubo = Knapsack(values, weights, weight_capacity, penalty).qubo

        self.assertTrue(terms_list_equality(knap_terms, knapsack_prob_qubo.terms))
        self.assertEqual(knap_weights, knapsack_prob_qubo.weights)
        self.assertEqual(knap_constant, knapsack_prob_qubo.constant)
        self.assertEqual(n_qubits, knapsack_prob_qubo.n)

    def test_knapsack_random_problem(self):
        """Test random instance method of Knapsack problem class"""

        rng = np.random.default_rng(1234)
        n_items = 5
        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))
        weight_capacity = int(
            rng.integers(np.min(weights) * n_items, np.max(weights) * n_items)
        )
        penalty = 2 * np.max(values)

        knap_manual = Knapsack(values, weights, weight_capacity, int(penalty)).qubo

        knap_random_instance = Knapsack.random_instance(n_items=n_items, seed=1234).qubo

        self.assertTrue(
            terms_list_equality(knap_manual.terms, knap_random_instance.terms)
        )
        self.assertEqual(knap_manual.weights, knap_random_instance.weights)
        self.assertEqual(knap_manual.constant, knap_random_instance.constant)
        self.assertEqual(knap_manual.n, knap_random_instance.n)

    def test_knapsack_random_problem_smallsize(self):
        """Test random instance method of Knapsack problem class"""

        rng = np.random.default_rng(1234)
        n_items = 3
        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))
        weight_capacity = int(
            rng.integers(np.min(weights) * n_items, np.max(weights) * n_items)
        )
        penalty = 2 * np.max(values)

        knap_manual = Knapsack(values, weights, weight_capacity, int(penalty)).qubo

        knap_random_instance = Knapsack.random_instance(n_items=n_items, seed=1234).qubo

        self.assertTrue(
            terms_list_equality(knap_manual.terms, knap_random_instance.terms)
        )
        self.assertEqual(knap_manual.weights, knap_random_instance.weights)
        self.assertEqual(knap_manual.constant, knap_random_instance.constant)
        self.assertEqual(knap_manual.n, knap_random_instance.n)

    def test_knapsack_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # values type-check
        weights = [1, 2]
        weight_capacity = 5
        penalty = 0.1
        values_list = [(1, 2), {"test": "oh", "test1": "oh"}, np.array([1, 2])]

        for each_values in values_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(each_values, weights, weight_capacity, penalty)
            self.assertEqual(
                "The input parameter, values, has to be a list", str(e.exception)
            )

        values_list = [["test", "oh"], [np.array(1), np.array(2)], [0.1, 0.5]]
        for each_values in values_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(each_values, weights, weight_capacity, penalty)
            self.assertEqual(
                "The elements in values list must be of type int.", str(e.exception)
            )

        # weights type-check
        values = [1, 2]
        weight_capacity = 5
        penalty = 0.1
        weights_list = [(1, 2), {"test": "oh", "test1": "oh"}, np.array([1, 2])]

        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, each_weights, weight_capacity, penalty)
            self.assertEqual(
                "The input parameter, weights, has to be a list", str(e.exception)
            )

        weights_list = [["test", "oh"], [np.array(1), np.array(2)], [0.1, 0.5]]
        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, each_weights, weight_capacity, penalty)
            self.assertEqual(
                "The elements in weights list must be of type int.", str(e.exception)
            )

        # weight capacity type-check
        values = [1, 2]
        weights = [1, 2]
        weight_capacity_list = [0.5, np.array(1), np.array(0.5), "oh"]
        penalty = 0.1

        for each_weight_capacity in weight_capacity_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, weights, each_weight_capacity, penalty)
            self.assertEqual(
                "The input parameter, weight_capacity, has to be of type int",
                str(e.exception),
            )

        with self.assertRaises(TypeError) as e:
            Knapsack(values, weights, -1, penalty)
        self.assertEqual(
            "The input parameter, weight_capacity, must be a positive integer greater than 0",
            str(e.exception),
        )

        # penalty capacity type-check
        values = [1, 2]
        weights = [1, 2]
        penalty_list = [np.array(1), np.array(0.5), "oh"]
        weight_capacity = 5

        for each_penalty in penalty_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, weights, weight_capacity, each_penalty)
            self.assertEqual(
                "The input parameter, penalty, has to be of type float or int",
                str(e.exception),
            )


class TestSlackFreeKnapsack(unittest.TestCase):
    """Tests for SlackFreeKnapsack class"""

    def test_slackfreeknapsack_terms_weights_constant(self):
        """Test that SlackFreeKnapsack creates the correct QUBO problem"""

        values = [2, 4, 3, 5]
        weights = [3, 6, 9, 1]
        weight_capacity = 15
        n_qubits = len(values)
        penalty = 2 * max(values)
        slknap_terms = [
            [0, 1],
            [0, 2],
            [0, 3],
            [1, 2],
            [1, 3],
            [2, 3],
            [0],
            [1],
            [2],
            [3],
        ]
        slknap_weights = [
            90.0,
            135.0,
            15.0,
            270.0,
            30.0,
            45.0,
            166.0,
            332.0,
            496.5,
            57.5,
        ]
        slknap_constant = 613.0

        slknapsack_prob_qubo = SlackFreeKnapsack(
            values, weights, weight_capacity, penalty
        ).qubo

        self.assertTrue(terms_list_equality(slknap_terms, slknapsack_prob_qubo.terms))
        self.assertEqual(slknap_weights, slknapsack_prob_qubo.weights)
        self.assertEqual(slknap_constant, slknapsack_prob_qubo.constant)
        self.assertEqual(n_qubits, slknapsack_prob_qubo.n)

    def test_slackfreeknapsack_random_problem(self):
        """Test random instance method of SlackFreeKnapsack problem class"""

        rng = np.random.default_rng(1234)
        n_items = 5
        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))
        weight_capacity = int(
            rng.integers(np.min(weights) * n_items, np.max(weights) * n_items)
        )
        penalty = 2 * np.max(values)

        slknap_manual = SlackFreeKnapsack(
            values, weights, weight_capacity, int(penalty)
        ).qubo

        slknap_random_instance = SlackFreeKnapsack.random_instance(
            n_items=n_items, seed=1234
        ).qubo

        self.assertTrue(
            terms_list_equality(slknap_manual.terms, slknap_random_instance.terms)
        )
        self.assertEqual(slknap_manual.weights, slknap_random_instance.weights)
        self.assertEqual(slknap_manual.constant, slknap_random_instance.constant)
        self.assertEqual(slknap_manual.n, slknap_random_instance.n)


if __name__ == "__main__":
    unittest.main()



import warnings
import unittest
import numpy as np
import networkx as nx
import copy
import inspect

from openqaoa import QAOA
import openqaoa.optimizers.pennylane as pl
from openqaoa.backends import create_device
from openqaoa.optimizers.training_vqa import PennyLaneOptimizer
from openqaoa.optimizers.pennylane.optimization_methods_pennylane import (
    AVAILABLE_OPTIMIZERS,
)
from openqaoa.derivatives.derivative_functions import derivative
from openqaoa.optimizers.logger_vqa import Logger
from openqaoa.qaoa_components import (
    create_qaoa_variational_params,
    QAOADescriptor,
    PauliOp,
    Hamiltonian,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends.qaoa_backend import get_qaoa_backend
from openqaoa.optimizers import get_optimizer
from openqaoa.derivatives.qfim import qfim as Qfim
from openqaoa.problems import QUBO, MinimumVertexCover


# list of optimizers to test, pennylane optimizers
list_optimizers = PennyLaneOptimizer.PENNYLANE_OPTIMIZERS

# create a problem
g = nx.circulant_graph(4, [1])
problem = MinimumVertexCover(g, field=1.0, penalty=10)
qubo_problem_1 = problem.qubo
qubo_problem_2 = QUBO.random_instance(5)
qubo_problem_3 = QUBO.random_instance(6)


class TestPennylaneOptimizers(unittest.TestCase):
    def setUp(self):
        self.log = Logger(
            {
                "func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "jac_func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "qfim_func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
            },
            {
                "root_nodes": ["func_evals", "jac_func_evals", "qfim_func_evals"],
                "best_update_structure": [],
            },
        )

        self.log.log_variables(
            {"func_evals": 0, "jac_func_evals": 0, "qfim_func_evals": 0}
        )

    def _run_method_workflows(self, method, problem):
        "helper function to run the test for any method using workflows"
        q = QAOA()
        q.set_classical_optimizer(method=method, maxiter=3, jac="finite_difference")
        q.compile(problem)
        q.optimize()

        assert len(q.result.most_probable_states["solutions_bitstrings"][0]) > 0

    def _run_method_manual(self, method, problem):
        "helper function tu run the test for any method using manual mode"

        cost_hamil = problem.hamiltonian
        mixer_hamil = X_mixer_hamiltonian(n_qubits=problem.n)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=2)
        device = create_device("local", "vectorized")
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor, device)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "ramp"
        )
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        # declare needed functions
        jac = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "gradient",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )
        qfim = Qfim(backend_obj_vectorized, variate_params, self.log)

        # Optimize
        vector_optimizer = get_optimizer(
            backend_obj_vectorized,
            variate_params,
            optimizer_dict={
                "method": method,
                "jac": jac,
                "maxiter": niter,
                "qfim": qfim,
                "optimizer_options": {"stepsize": stepsize},
            },
        )
        vector_optimizer()

        # saving the results
        results = vector_optimizer.qaoa_result

        assert len(results.most_probable_states["solutions_bitstrings"][0]) == problem.n

    def test_pennylane_optimizers_workflows(self):
        "function to run the tests for pennylane optimizers, workflows"

        i = 0
        for problem in [qubo_problem_3, qubo_problem_2, qubo_problem_1]:
            for opt in list_optimizers:
                self._run_method_workflows(opt, problem)
                i += 1

        assert i == 3 * len(list_optimizers)

    def test_pennylane_optimizers_manual(self):
        "function to run the tests for pennylane optimizers, manual mode"

        i = 0
        for problem in [qubo_problem_3, qubo_problem_2, qubo_problem_1]:
            for opt in list_optimizers:
                self._run_method_manual(opt, problem)
                i += 1

        assert i == 3 * len(list_optimizers)

    def _pennylane_step(self, params_array, cost, optimizer, method, jac, qfim):
        "helper function to run a setp of the pennylane optimizer"
        params_array = pl.numpy.array(params_array, requires_grad=True)
        if method in [
            "pennylane_adagrad",
            "pennylane_adam",
            "pennylane_vgd",
            "pennylane_momentum",
            "pennylane_nesterov_momentum",
            "pennylane_rmsprop",
        ]:
            x, y = optimizer.step_and_cost(cost, params_array, grad_fn=jac)
        if method in ["pennylane_rotosolve"]:
            x, y = optimizer.step_and_cost(
                cost,
                params_array,
                nums_frequency={"params": {(i,): 1 for i in range(params_array.size)}},
            )
        if method in ["pennylane_spsa"]:
            x, y = optimizer.step_and_cost(cost, params_array)

        return x, y

    def test_step_and_cost(self):
        "function to run the tests for steps of pennylane optimizers"

        # define some problem
        cost_hamil = Hamiltonian(
            [
                PauliOp("ZZ", (0, 1)),
                PauliOp("ZZ", (1, 2)),
                PauliOp("ZZ", (0, 3)),
                PauliOp("Z", (2,)),
                PauliOp("Z", (1,)),
            ],
            [1, 1.1, 1.5, 2, -0.8],
            0.8,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=4)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=2)
        device = create_device("local", "vectorized")
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor, device)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "ramp"
        )
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        # declare needed functions
        jac = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "gradient",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )
        qfim = Qfim(backend_obj_vectorized, variate_params, self.log)

        def cost(params):
            variate_params.update_from_raw(params)
            return np.real(backend_obj_vectorized.expectation(variate_params))

        i = 0
        for method in list_optimizers:
            pennylane_method = method

            # copy the parameters
            x0 = copy.deepcopy(variate_params.raw().copy())

            # Optimize with the implemented optimizer in OpenQAOA
            vector_optimizer = get_optimizer(
                backend_obj_vectorized,
                variate_params,
                optimizer_dict={
                    "method": method,
                    "jac": jac,
                    "maxiter": niter,
                    "qfim": qfim,
                    "optimizer_options": {"stepsize": stepsize},
                },
            )
            vector_optimizer()

            # formatting the data
            y_opt = vector_optimizer.qaoa_result.intermediate["cost"][1:4]
            if pennylane_method in ["pennylane_rotosolve"]:
                y_opt = vector_optimizer.qaoa_result.intermediate["cost"][4:40:12]

            # get optimizer to try
            optimizer = AVAILABLE_OPTIMIZERS[pennylane_method]
            # get optimizer arguments
            arguments = inspect.signature(optimizer).parameters.keys()

            # check if stepsize is in the optimizer arguments
            options = {}
            if "stepsize" in arguments:
                options["stepsize"] = stepsize
            if "maxiter" in arguments:
                options["maxiter"] = niter

            # pass the argument to the optimizer
            optimizer = optimizer(**options)

            # reinitialize variables
            variate_params.update_from_raw(x0)
            x0 = variate_params.raw().copy()
            y0 = cost(x0)

            # compute steps (depends on the optimizer)
            x1, y1 = self._pennylane_step(
                x0, cost, optimizer, pennylane_method, jac, qfim
            )
            x2, y2 = self._pennylane_step(
                x1, cost, optimizer, pennylane_method, jac, qfim
            )
            x3, y3 = self._pennylane_step(
                x2, cost, optimizer, pennylane_method, jac, qfim
            )

            # list of results
            y = [y1, y2, y3]

            # check that the results are ok
            if pennylane_method in ["pennylane_spsa"]:
                assert np.sum(np.abs(np.array(y)) >= 0) == 3
            else:
                for yi, y_opt_i in zip(y, y_opt):
                    assert np.isclose(yi, y_opt_i, rtol=0.001, atol=0.001)

            i += 1

        assert i == len(list_optimizers)


if __name__ == "__main__":
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=PendingDeprecationWarning)
        unittest.main()



from typing import List, Union
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np

from .variational_baseparams import QAOAVariationalBaseParams
from ..ansatz_constructor import QAOADescriptor
from ..ansatz_constructor.baseparams import shapedArray, _is_iterable_empty


class QAOAVariationalStandardParams(QAOAVariationalBaseParams):
    r"""
    QAOA parameters that implement a state preparation circuit with

    .. math::

        e^{-i \beta_p H_0}
        e^{-i \gamma_p H_c}
        \cdots
        e^{-i \beta_0 H_0}
        e^{-i \gamma_0 H_c}

    This corresponds to the parametrization used by Farhi in his
    original paper [https://arxiv.org/abs/1411.4028]

    Parameters
    ----------
    qaoa_descriptor:
        QAOADescriptor object containing circuit instructions
    betas:
        List of p betas
    gammas:
        List of p gammas

    Attributes
    ----------
    betas: np.array
        1D array with the betas from above
    gammas: np.array
        1D array with the gamma from above
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        betas: List[Union[float, int]],
        gammas: List[Union[float, int]],
    ):
        # setup reg, qubits_singles and qubits_pairs
        super().__init__(qaoa_descriptor)
        self.betas = np.array(betas)
        self.gammas = np.array(gammas)

    def __repr__(self):
        string = "Standard Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "Variational Parameters:\n"
        string += "\tbetas: " + str(self.betas) + "\n"
        string += "\tgammas: " + str(self.gammas) + "\n"
        return string

    def __len__(self):
        return self.p * 2

    @shapedArray
    def betas(self):
        return self.p

    @shapedArray
    def gammas(self):
        return self.p

    @property
    def mixer_1q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_1q_coeffs)

    @property
    def mixer_2q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_2q_coeffs)

    @property
    def cost_1q_angles(self):
        return 2 * np.outer(self.gammas, self.cost_1q_coeffs)

    @property
    def cost_2q_angles(self):
        return 2 * np.outer(self.gammas, self.cost_2q_coeffs)

    def update_from_raw(self, new_values):
        # overwrite self.betas with new ones
        self.betas = np.array(new_values[0 : self.p])
        new_values = new_values[self.p :]  # cut betas from new_values
        self.gammas = np.array(new_values[0 : self.p])
        new_values = new_values[self.p :]

        if len(new_values) != 0:
            raise RuntimeWarning(
                "Incorrect dimension specified for new_values"
                "to construct the new betas and new gammas"
            )

    def raw(self):
        raw_data = np.concatenate((self.betas, self.gammas))
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls, qaoa_descriptor: QAOADescriptor, time: float = None
    ):
        """
        Returns
        -------
        StandardParams
            A ``StandardParams`` object with parameters according
            to a linear ramp schedule for the Hamiltonian specified by register, terms, weights.
        """
        p = qaoa_descriptor.p

        if time is None:
            time = float(0.7 * p)
        # create evenly spaced timelayers at the centers of p intervals
        dt = time / p
        # fill betas, gammas_singles and gammas_pairs
        betas = np.linspace(
            (dt / time) * (time * (1 - 0.5 / p)), (dt / time) * (time * 0.5 / p), p
        )
        gammas = betas[::-1]
        # wrap it all nicely in a qaoa_parameters object
        params = cls(qaoa_descriptor, betas, gammas)

        return params

    @classmethod
    def random(cls, qaoa_descriptor: QAOADescriptor, seed: int = None):
        """
        Returns
        -------
        StandardParams
            Randomly initialised ``StandardParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        betas = np.random.uniform(0, np.pi, qaoa_descriptor.p)
        gammas = np.random.uniform(0, np.pi, qaoa_descriptor.p)

        params = cls(qaoa_descriptor, betas, gammas)

        return params

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor):
        """
        Initialise Standard Variational params with empty arrays
        """
        p = qaoa_descriptor.p
        betas = np.empty(p)
        gammas = np.empty(p)

        return cls(qaoa_descriptor, betas, gammas)

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(self.betas, label="betas", marker="s", ls="", **kwargs)
        ax.plot(self.gammas, label="gammas", marker="^", ls="", **kwargs)
        ax.set_xlabel("p", fontsize=12)
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.legend()

    def convert_to_ext(self, args_std):
        """
        Method that converts a list of parameters in the standard parametrisation
        form (args_std) to an equivalent list of parameters in the extended parametrisation form.

        PARAMETERS
        ----------
        args_std :
            Parameters (a list of float) in the standard parametrisation form.

        RETURNS
        -------
        args_ext:
            Parameters (a list of float) in the extended parametrisation form.

        """

        terms_lst = [
            len(self.mixer_1q_coeffs),
            len(self.mixer_2q_coeffs),
            len(self.cost_1q_coeffs),
            len(self.cost_2q_coeffs),
        ]
        terms_lst_p = np.repeat(terms_lst, [self.p] * len(terms_lst))
        args_ext = []
        for i in range(4):  # 4 types of terms
            for j in range(self.p):
                for k in range(terms_lst_p[i * self.p + j]):
                    if i < 2:
                        args_ext.append(args_std[j])
                    else:
                        args_ext.append(args_std[j + int(len(args_std) / 2)])

        return args_ext


class QAOAVariationalStandardWithBiasParams(QAOAVariationalBaseParams):
    r"""
    QAOA parameters that implement a state preparation circuit with

    .. math::

        e^{-i \beta_p H_0}
        e^{-i \gamma_{\textrm{singles}, p} H_{c, \textrm{singles}}}
        e^{-i \gamma_{\textrm{pairs}, p} H_{c, \textrm{pairs}}}
        \cdots
        e^{-i \beta_0 H_0}
        e^{-i \gamma_{\textrm{singles}, 0} H_{c, \textrm{singles}}}
        e^{-i \gamma_{\textrm{pairs}, 0} H_{c, \textrm{pairs}}}

    where the cost hamiltonian is split into :math:`H_{c, \textrm{singles}}`
    the bias terms, that act on only one qubit, and
    :math:`H_{c, \textrm{pairs}}` the coupling terms, that act on two qubits.

    Parameters
    ----------
    qaoa_descriptor:
        QAOADescriptor object containing circuit instructions
    betas:
        List of p betas
    gammas_singles:
        List of p gammas_singles
    gammas_pairs:
        List of p gammas_pairs

    Attributes
    ----------
    betas: np.array
        A 1D array containing the betas from above for each timestep
    gammas_pairs: np.array
        A 1D array containing the gammas_singles from above for each timestep
    gammas_singles: np.array
        A 1D array containing the gammas_pairs from above for each timestep
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        betas: List[Union[float, int]],
        gammas_singles: List[Union[float, int]],
        gammas_pairs: List[Union[float, int]],
    ):
        super().__init__(qaoa_descriptor)
        if not self.cost_1q_coeffs or not self.cost_2q_coeffs:
            raise RuntimeError(
                f"Please choose {type(self).__name__} parameterisation for "
                "problems containing both Cost One-Qubit and Two-Qubit terms"
            )

        self.betas = np.array(betas)
        self.gammas_singles = np.array(gammas_singles)
        self.gammas_pairs = np.array(gammas_pairs)

    def __repr__(self):
        string = "Standard with Bias Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "Variational Parameters:\n"
        string += "\tbetas: " + str(self.betas) + "\n"
        string += "\tgammas_singles: " + str(self.gammas_singles) + "\n"
        string += "\tgammas_pairs: " + str(self.gammas_pairs) + "\n"
        return string

    def __len__(self):
        return self.p * 3

    @shapedArray
    def betas(self):
        return self.p

    @shapedArray
    def gammas_singles(self):
        return self.p

    @shapedArray
    def gammas_pairs(self):
        return self.p

    @property
    def mixer_1q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_1q_coeffs)

    @property
    def mixer_2q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_2q_coeffs)

    @property
    def cost_1q_angles(self):
        return 2 * np.outer(self.gammas_singles, self.cost_1q_coeffs)

    @property
    def cost_2q_angles(self):
        return 2 * np.outer(self.gammas_pairs, self.cost_2q_coeffs)

    def update_from_raw(self, new_values):
        # overwrite self.betas with new ones
        self.betas = np.array(new_values[0 : self.p])
        new_values = new_values[self.p :]  # cut betas from new_values
        self.gammas_singles = np.array(new_values[0 : self.p])
        new_values = new_values[self.p :]
        self.gammas_pairs = np.array(new_values[0 : self.p])
        new_values = new_values[self.p :]

        if len(new_values) != 0:
            raise RuntimeWarning(
                "Incorrect dimension specified for new_values"
                "to construct the new betas and new gammas"
            )

    def raw(self):
        raw_data = np.concatenate((self.betas, self.gammas_singles, self.gammas_pairs))
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls, qaoa_descriptor: QAOADescriptor, time: float = None
    ):
        """
        Returns
        -------
        StandardParams
            A ``StandardParams`` object with parameters according
            to a linear ramp schedule for the Hamiltonian specified by register, terms, weights.
        """
        p = qaoa_descriptor.p

        if time is None:
            time = float(0.7 * p)
        # create evenly spaced timelayers at the centers of p intervals
        dt = time / p
        # fill betas, gammas_singles and gammas_pairs
        betas = np.linspace(
            (dt / time) * (time * (1 - 0.5 / p)), (dt / time) * (time * 0.5 / p), p
        )
        gammas_singles = betas[::-1]
        gammas_pairs = betas[::-1]

        params = cls(qaoa_descriptor, betas, gammas_singles, gammas_pairs)

        return params

    @classmethod
    def random(cls, qaoa_descriptor: QAOADescriptor, seed: int = None):
        """
        Returns
        -------
        StandardParams
            Randomly initialised ``StandardParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        betas = np.random.uniform(0, np.pi, qaoa_descriptor.p)
        gammas_singles = np.random.uniform(0, np.pi, qaoa_descriptor.p)
        gammas_pairs = np.random.uniform(0, np.pi, qaoa_descriptor.p)

        params = cls(qaoa_descriptor, betas, gammas_singles, gammas_pairs)
        return params

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor):
        """
        Initialise Standard Variational params with empty arrays
        """
        p = qaoa_descriptor.p
        betas = np.empty(p)
        gammas_singles = np.empty(p)
        gammas_pairs = np.empty(p)

        return cls(qaoa_descriptor, betas, gammas_singles, gammas_pairs)

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(self.betas, label="betas", marker="s", ls="", **kwargs)
        if not _is_iterable_empty(self.gammas_singles):
            ax.plot(
                self.gammas_singles, label="gammas_singles", marker="^", ls="", **kwargs
            )
        if not _is_iterable_empty(self.gammas_pairs):
            ax.plot(
                self.gammas_pairs, label="gammas_pairs", marker="v", ls="", **kwargs
            )
        ax.set_xlabel("p")
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        # ax.grid(linestyle='--')
        ax.legend()



import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from typing import Union
from docplex.mp.model import Model

from .problem import Problem
from .converters import FromDocplex2IsingModel
from .qubo import QUBO


class SK(Problem):
    """
    Creates an instance of the Sherrington Kirkpatrick.
    https://en.wikipedia.org/wiki/Spin_glass

    Parameters
    ----------
    G: networkx.Graph
        Networkx graph of the problem


    Returns
    -------
        An instance of the Sherrington Kirkpatrick Problem.
    """

    __name__ = "sherrington_kirkpatrick"

    def __init__(self, G):
        self.G = G

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):
        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the SK problem, whose graph is random following the Erdos-Renyi model.

        Parameters
        ----------
        **kwargs:

            n_nodes: int
                The number of nodes (spins) in the graph.
            mu: float
                Mean value of the distribution for the J_{ij} values.
            sigma: float
                Standard deviation of the distribution for the J_{ij} values.
            seed: int
                Seed to be used for the random case.

        Returns
        -------
            A random instance of the SK problem.
        """
        n_nodes = kwargs.get("n_nodes")
        mu = kwargs.get("mu", 0)
        sigma = kwargs.get("sigma", 1)
        seed = kwargs.get("seed", None)
        rng = np.random.default_rng(seed)

        G = nx.Graph()
        G.add_weighted_edges_from(
            [
                [i, j, round(rng.normal(mu, sigma), 3)]
                for i in range(n_nodes)
                for j in range(i + 1, n_nodes)
            ]
        )

        return SK(G)

    @property
    def docplex_model(self):
        mdl = Model("SK_Problem")
        num_spins = self.G.number_of_nodes()

        x = {i: mdl.binary_var(name=f"x_{i}") for i in range(num_spins)}

        energy = -1 * mdl.sum(
            (2 * x[i] - 1) * (2 * x[j] - 1) * self.G[i][j]["weight"]
            for i, j in self.G.edges()
        )

        mdl.minimize(energy)

        return mdl

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        cplex_model = self.docplex_model
        qubo_docplex = FromDocplex2IsingModel(cplex_model).ising_model

        return QUBO(
            self.G.number_of_nodes(),
            qubo_docplex.terms + [[]],
            qubo_docplex.weights + [qubo_docplex.constant],
            self.problem_instance,
        )

    def classical_solution(self, string: bool = False):
        """
        Solves the SK problem and returns the CPLEX solution.

        Returns
        -------
        solution : dict
            A dictionary containing the solution found with spin values.
        """
        model = self.docplex_model
        model.solve()
        status = model.solve_details.status
        if status != "integer optimal solution":
            print(status)
        if string:
            return "".join(
                str(round(model.solution.get_value(var)))
                for var in model.iter_binary_vars()
            )
        else:
            return {
                var.name: model.solution.get_value(var)
                for var in model.iter_binary_vars()
            }

    def plot_solution(self, solution: Union[str, dict], ax=None):
        """
        Plots the solution of the SK problem.

        Parameters
        ----------
        solution : dict
            A dictionary containing the solution with spin values.

        Returns
        -------
        fig : matplotlib.pyplot.Figure()
            The graph visualization of the solution.
        """
        colors = ["#5EB1EB", "#F29D55"]
        if isinstance(solution, str):
            sol = {}
            for n, var in enumerate(self.docplex_model.iter_binary_vars()):
                sol[var.name] = int(solution[n])
            solution = sol
        if ax is None:
            _, ax = plt.subplots()

        pos = nx.circular_layout(self.G)
        nx.draw(
            self.G,
            pos=pos,
            edgecolors="black",
            node_color=[colors[round(solution[spin])] for spin in solution],
            labels={n: n for n in range(len(solution))},
            ax=ax,
            edge_color="#0B2340",
        )
        edge_labels = nx.get_edge_attributes(self.G, "weight")

        nx.draw_networkx_edge_labels(self.G, pos=pos, edge_labels=edge_labels, ax=ax)
        ax.set_title("SK Problem Solution")
        ax.plot([], [], color=colors[0], label=r"$S_{-1}$", marker="o", linewidth=0)
        ax.plot([], [], color=colors[1], label=r"$S_{+1}$", marker="o", linewidth=0)
        ax.legend()
